import shutil
import unittest
from pathlib import Path

import runhouse as rh
from ray import cloudpickle as pickle

TEST_FOLDER_PATH = Path.cwd() / "tests_tmp"

DATA_STORE_BUCKET = "runhouse-folder"
DATA_STORE_PATH = f"/{DATA_STORE_BUCKET}/folder-tests"


def setup():
    from pathlib import Path

    # Create buckets in S3 and GCS
    from runhouse.rns.api_utils.utils import create_gcs_bucket, create_s3_bucket

    create_s3_bucket(DATA_STORE_BUCKET)
    create_gcs_bucket(DATA_STORE_BUCKET)

    # Create local dir with files to upload to cluster, buckets, etc.
    TEST_FOLDER_PATH.mkdir(parents=True, exist_ok=True)
    for i in range(3):
        output_file = Path(f"{TEST_FOLDER_PATH}/sample_file_{i}.txt")
        output_file.write_text(f"file{i}")


def delete_local_folder(path):
    shutil.rmtree(path)


def fs_str_rh_fn(folder):
    return folder._fs_str


def test_github_folder():
    gh_folder = rh.folder(
        path="/", system="github", data_config={"org": "pytorch", "repo": "pytorch"}
    )
    assert gh_folder.ls()


# ----------------- Run tests -----------------


def test_from_cluster():
    cluster = rh.cluster(name="^rh-cpu").up_if_not()
    rh.folder(path=str(Path.cwd())).to(cluster, path="~/my_new_tests_folder")
    tests_folder = rh.folder(system=cluster, path="~/my_new_tests_folder")
    assert "my_new_tests_folder/test_folder.py" in tests_folder.ls()


def test_to_cluster_attr():
    cluster = rh.cluster(name="^rh-cpu").up_if_not()
    local_folder = rh.folder(path=TEST_FOLDER_PATH)
    cluster_folder = local_folder.to(system=cluster)
    assert isinstance(cluster_folder.system, rh.Cluster)
    assert cluster_folder._fs_str == "ssh"

    s3_folder = rh.folder(path=TEST_FOLDER_PATH, system="s3")
    cluster_folder_s3 = s3_folder.to(system=cluster)
    assert isinstance(cluster_folder_s3.system, rh.Cluster)
    assert cluster_folder_s3._fs_str == "ssh"


def test_create_and_save_data_to_s3_folder():
    data = list(range(50))
    s3_folder = rh.folder(path=DATA_STORE_PATH, system="s3")
    s3_folder.mkdir()
    s3_folder.put({"test_data.py": pickle.dumps(data)}, overwrite=True)

    assert s3_folder.exists_in_system()


def test_read_data_from_existing_s3_folder():
    # Note: Uses folder created above
    s3_folder = rh.folder(path=DATA_STORE_PATH, system="s3")
    fss_file: "fsspec.core.OpenFile" = s3_folder.open(name="test_data.py")
    with fss_file as f:
        data = pickle.load(f)

    assert data == list(range(50))


def test_create_and_delete_folder_from_s3():
    s3_folder = rh.folder(name=DATA_STORE_PATH, system="s3")
    s3_folder.mkdir()

    s3_folder.delete_configs()

    assert not s3_folder.exists_in_system()


def test_folder_attr_on_cluster():
    c = rh.cluster("^rh-cpu").up_if_not()
    cluster_folder = rh.folder(path=TEST_FOLDER_PATH).to(system=c)
    fs_str_cluster = rh.function(fn=fs_str_rh_fn).to(system=c)
    fs_str = fs_str_cluster(cluster_folder)
    assert fs_str == "file"


def test_cluster_tos():
    tests_folder = rh.folder(path=str(Path.cwd()))

    c = rh.cluster("^rh-cpu").up_if_not()
    tests_folder = tests_folder.to(system=c)
    assert "test_folder.py" in tests_folder.ls(full_paths=False)

    # to local
    local = tests_folder.to("here", path=TEST_FOLDER_PATH)
    assert "test_folder.py" in local.ls(full_paths=False)

    # to s3
    s3 = tests_folder.to("s3")
    assert "test_folder.py" in s3.ls(full_paths=False)

    delete_local_folder(TEST_FOLDER_PATH)
    s3.delete_in_system()

    # to gcs
    gcs = tests_folder.to("gs")
    try:
        assert "test_folder.py" in gcs.ls(full_paths=False)
        gcs.delete_in_system()

    except:
        bucket_name = gcs.bucket_name_from_path(gcs.path)
        print(
            f"Permissions to gs bucket ({bucket_name}) may not be fully enabled "
            f"on the cluster {c.name}. For now please manually enable them directly on the cluster. "
            f"See https://cloud.google.com/sdk/gcloud/reference/auth/login"
        )


def test_local_and_cluster():
    # Local to cluster
    local_folder = rh.folder(path=TEST_FOLDER_PATH)
    local_folder.mkdir()
    local_folder.put({f"sample_file_{i}.txt": f"file{i}".encode() for i in range(3)})
    c = rh.cluster("^rh-cpu").up_if_not()
    cluster_folder = local_folder.to(system=c)
    assert "sample_file_0.txt" in cluster_folder.ls(full_paths=False)

    # Cluster to local
    tmp_path = Path.cwd() / "tmp_from_cluster"
    local_from_cluster = cluster_folder.to("here", path=tmp_path)
    assert "sample_file_0.txt" in local_from_cluster.ls(full_paths=False)

    delete_local_folder(tmp_path)


def test_local_and_s3():
    # Local to S3
    local_folder = rh.folder(path=TEST_FOLDER_PATH)
    s3_folder = local_folder.to(system="s3")
    assert "sample_file_0.txt" in s3_folder.ls(full_paths=False)

    # S3 to local
    tmp_path = Path.cwd() / "tmp_from_s3"
    tmp_path.mkdir(exist_ok=True)

    local_from_s3 = s3_folder.to("here", path=tmp_path)
    assert "sample_file_0.txt" in local_from_s3.ls(full_paths=False)

    delete_local_folder(tmp_path)
    s3_folder.delete_in_system()


def test_local_and_gcs():
    # Local to GCS
    local_folder = rh.folder(path=TEST_FOLDER_PATH)
    gcs_folder = local_folder.to(system="gs")
    assert "sample_file_0.txt" in gcs_folder.ls(full_paths=False)

    # GCS to local
    tmp_path = Path.cwd() / "tmp_from_gcs"
    tmp_path.mkdir(exist_ok=True)

    local_from_gcs = gcs_folder.to("here", path=tmp_path)
    assert "sample_file_0.txt" in local_from_gcs.ls(full_paths=False)

    delete_local_folder(tmp_path)
    gcs_folder.delete_in_system()


def test_cluster_and_s3():
    # Local to cluster
    local_folder = rh.folder(path=TEST_FOLDER_PATH)
    c = rh.cluster("^rh-cpu").up_if_not()
    cluster_folder = local_folder.to(system=c)
    assert "sample_file_0.txt" in cluster_folder.ls(full_paths=False)

    # Cluster to S3
    s3_folder = cluster_folder.to(system="s3")
    assert "sample_file_0.txt" in s3_folder.ls(full_paths=False)

    # S3 to cluster
    cluster_from_s3 = s3_folder.to(system=c)
    assert "sample_file_0.txt" in cluster_from_s3.ls(full_paths=False)

    s3_folder.delete_in_system()


@unittest.skip("requires GCS setup")
def test_cluster_and_gcs():
    # Local to cluster
    local_folder = rh.folder(path=TEST_FOLDER_PATH)
    c = rh.cluster("^rh-cpu").up_if_not()

    # Make sure we have gsutil and gcloud on the cluster - needed for copying the package + authenticating
    c.install_packages(["gsutil"])

    # TODO [JL] might be necessary to install gcloud on the cluster
    # c.run(['sudo snap install google-cloud-cli --classic'])

    cluster_folder = local_folder.to(system=c)
    assert "sample_file_0.txt" in cluster_folder.ls(full_paths=False)

    # Cluster to GCS
    gcs_folder = cluster_folder.to(system="gs")
    try:
        assert "sample_file_0.txt" in gcs_folder.ls(full_paths=False)
        # GCS to cluster
        cluster_from_gcs = gcs_folder.to(system=c)
        assert "sample_file_0.txt" in cluster_from_gcs.ls(full_paths=False)

        gcs_folder.delete_in_system()

    except:
        # TODO [JL] automate gcloud access on the cluster for writing to GCS bucket
        bucket_name = gcs_folder.bucket_name_from_path(gcs_folder.path)
        raise PermissionError(
            f"Permissions to gs bucket ({bucket_name}) may not be fully enabled "
            f"on the cluster {c.name}. For now please manually enable them directly on the cluster. "
            f"See https://cloud.google.com/sdk/gcloud/reference/auth/login"
        )


def test_s3_and_s3():
    # Local to S3
    local_folder = rh.folder(path=TEST_FOLDER_PATH)
    s3_folder = local_folder.to(system="s3")
    assert "sample_file_0.txt" in s3_folder.ls(full_paths=False)

    # from one s3 folder to another s3 folder
    new_s3_folder = s3_folder.to(system="s3")
    assert "sample_file_0.txt" in new_s3_folder.ls(full_paths=False)

    s3_folder.delete_in_system()
    new_s3_folder.delete_in_system()


def test_gcs_and_gcs():
    # Local to GCS
    local_folder = rh.folder(path=TEST_FOLDER_PATH)
    gcs_folder = local_folder.to(system="gs")
    assert "sample_file_0.txt" in gcs_folder.ls(full_paths=False)

    # from one gcs folder to another gcs folder
    new_gcs_folder = gcs_folder.to(system="gs")
    assert "sample_file_0.txt" in new_gcs_folder.ls(full_paths=False)

    gcs_folder.delete_in_system()
    new_gcs_folder.delete_in_system()


def test_s3_and_gcs():
    # Local to S3
    local_folder = rh.folder(path=TEST_FOLDER_PATH)
    s3_folder = local_folder.to(system="s3")
    assert "sample_file_0.txt" in s3_folder.ls(full_paths=False)

    # *** NOTE: transfers between providers are only supported at the bucket level at the moment (not directory) ***

    # S3 to GCS
    s3_folder_to_gcs = s3_folder.to(system="gs")
    assert s3_folder_to_gcs.ls(full_paths=False)

    s3_folder.delete_in_system()


@unittest.skip("requires GCS setup")
def test_gcs_and_s3():
    # Local to GCS
    local_folder = rh.folder(path=TEST_FOLDER_PATH)
    gcs_folder = local_folder.to(system="gs")
    assert "sample_file_0.txt" in gcs_folder.ls(full_paths=False)

    # *** NOTE: transfers between providers are only supported at the bucket level at the moment (not directory) ***

    # GCS to S3
    gcs_folder_to_s3 = gcs_folder.to(system="s3")
    assert gcs_folder_to_s3.ls(full_paths=False)

    gcs_folder.delete_in_system()


def test_s3_folder_uploads_and_downloads():
    # NOTE: you can specify a specific path like this:
    # test_folder = rh.folder(path='/runhouse/my-folder', system='gs')

    s3_folder = rh.folder(system="s3")
    s3_folder.upload(src=str(TEST_FOLDER_PATH))

    assert s3_folder.exists_in_system()

    downloaded_path_folder = str(Path.cwd() / "downloaded_s3")
    s3_folder.download(dest=downloaded_path_folder)

    assert Path(downloaded_path_folder).exists()

    # remove folder in s3
    s3_folder.delete_in_system()
    assert not s3_folder.exists_in_system()


def test_cluster_and_cluster():
    # Local to cluster 1
    local_folder = rh.folder(path=TEST_FOLDER_PATH)
    c1 = rh.cluster("^rh-cpu").up_if_not()

    # Upload sky secrets to cluster - required when syncing over the folder from c1 to c2
    c1.send_secrets(providers=["sky"])

    cluster_folder_1 = local_folder.to(system=c1)
    assert "sample_file_0.txt" in cluster_folder_1.ls(full_paths=False)

    # Cluster 1 to cluster 2
    c2 = rh.cluster(name="test-byo-cluster").up_if_not()
    cluster_folder_2 = cluster_folder_1.to(system=c2, path=cluster_folder_1.path)
    assert "sample_file_0.txt" in cluster_folder_2.ls(full_paths=False)


def test_s3_sharing():
    s3_folder = rh.folder(
        name="my-s3-shared-folder", path=DATA_STORE_PATH, system="s3"
    ).save()
    assert s3_folder.ls(full_paths=False)

    s3_folder.share(
        users=["donny@run.house", "josh.lewittes@gmail.com"],
        access_type="read",
        notify_users=False,
    )

    assert s3_folder.ls(full_paths=False)


def test_load_shared_folder():
    from runhouse import Folder

    my_folder = Folder.from_name("@/my-s3-shared-folder")
    folder_contents = my_folder.ls()
    assert folder_contents


if __name__ == "__main__":
    setup()
    unittest.main()



import logging
import subprocess
from pathlib import Path

from runhouse import rh_config
from .package import Package


class GitPackage(Package):
    RESOURCE_TYPE = "package"

    def __init__(
        self,
        name: str = None,
        git_url: str = None,
        install_method: str = None,
        install_args: str = None,
        revision: str = None,
        dryrun: bool = False,
        **kwargs,  # We have this here to ignore extra arguments when calling from from_config
    ):
        """
        Runhouse Github Package resource.

        Example:
            >>> rh.GitPackage(git_url='https://github.com/runhouse/runhouse.git',
            >>>               install_method='pip', revision='v0.0.1')
        """
        super().__init__(
            name=name,
            dryrun=dryrun,
            install_method=install_method,
            install_target="./" + git_url.split("/")[-1].replace(".git", ""),
            install_args=install_args,
        )
        self.git_url = git_url
        self.revision = revision

    @property
    def config_for_rns(self):
        # If the package is just a simple Package.from_string string, no
        # need to store it in rns, just give back the string.
        # if self.install_method in ['pip', 'conda', 'git']:
        #     return f'{self.install_method}:{self.name}'
        config = super().config_for_rns
        self.save_attrs_to_config(config, ["git_url", "revision"])
        return config

    def __str__(self):
        if self.name:
            return f"GitPackage: {self.name}"
        return f"GitPackage: {self.git_url}@{self.revision}"

    def install(self):
        # Clone down the repo
        if not Path(self.install_target).exists():
            logging.info(f"Cloning: git clone {self.git_url}")
            subprocess.check_call(["git", "clone", self.git_url])
        else:
            logging.info(f"Pulling: git -C {self.install_target} fetch {self.git_url}")
            subprocess.check_call(
                f"git -C {self.install_target} fetch {self.git_url}".split(" ")
            )
        # Checkout the revision
        if self.revision:
            logging.info(f"Checking out revision: git checkout {self.revision}")
            subprocess.check_call(
                ["git", "-C", self.install_target, "checkout", self.revision]
            )
        # Use super to install the package
        super().install()

    @staticmethod
    def from_config(config: dict, dryrun=False):
        return GitPackage(**config, dryrun=dryrun)


def git_package(
    name=None,
    git_url: str = None,
    revision: str = None,
    install_method: str = None,
    install_str: str = None,
    dryrun=False,
):
    config = rh_config.rns_client.load_config(name)
    config["name"] = name or config.get("rns_address", None) or config.get("name")

    config["install_method"] = install_method or config.get("install_method", "local")
    config["install_args"] = install_str or config.get("install_args")
    config["revision"] = revision or config.get("revision")
    if git_url is not None:
        if not git_url.endswith(".git"):
            git_url += ".git"
        config["git_url"] = git_url

    new_package = GitPackage.from_config(config, dryrun=dryrun)

    return new_package



from enum import Enum

from runhouse.rns.secrets.aws_secrets import AWSSecrets
from runhouse.rns.secrets.azure_secrets import AzureSecrets
from runhouse.rns.secrets.gcp_secrets import GCPSecrets
from runhouse.rns.secrets.github_secrets import GitHubSecrets
from runhouse.rns.secrets.huggingface_secrets import HuggingFaceSecrets
from runhouse.rns.secrets.lambda_secrets import LambdaSecrets
from runhouse.rns.secrets.sky_secrets import SkySecrets
from runhouse.rns.secrets.ssh_secrets import SSHSecrets


class Providers(Enum):
    AWS = AWSSecrets()
    AZURE = AzureSecrets()
    GCP = GCPSecrets()
    HUGGINGFACE = HuggingFaceSecrets()
    LAMBDA = LambdaSecrets()
    SKY = SkySecrets()
    SSH = SSHSecrets()
    GITHUB = GitHubSecrets()



import copy
import inspect
import json
import logging
import os
import re
from pathlib import Path
from typing import Callable, List, Optional, Tuple, Union

import requests

from runhouse import rh_config
from runhouse.rns.api_utils.resource_access import ResourceAccess
from runhouse.rns.api_utils.utils import is_jsonable, load_resp_content, read_resp_data
from runhouse.rns.hardware import Cluster
from runhouse.rns.packages import git_package, Package

from runhouse.rns.resource import Resource
from runhouse.rns.run_module_utils import call_fn_by_type, get_fn_by_name

logger = logging.getLogger(__name__)


class Function(Resource):
    RESOURCE_TYPE = "function"
    DEFAULT_ACCESS = "write"

    def __init__(
        self,
        fn_pointers: Tuple[str, str, str],
        system: Optional[Cluster] = None,
        name: Optional[str] = None,
        reqs: Optional[List[str]] = None,
        setup_cmds: Optional[List[str]] = None,
        dryrun: bool = False,
        access: Optional[str] = None,
        **kwargs,  # We have this here to ignore extra arguments when calling from from_config
    ):
        """
        Runhouse Function object. It comprises of the entrypoint, system/cluster,
        and dependencies necessary to run the service.

        .. note::
                To create a Function, please use the factory method :func:`function`.
        """
        self.fn_pointers = fn_pointers
        self.system = system
        self.reqs = reqs
        self.setup_cmds = setup_cmds or []
        self.access = access or self.DEFAULT_ACCESS
        self.dryrun = dryrun
        super().__init__(name=name, dryrun=dryrun)

        if not self.dryrun and self.system and self.access in ["write", "read"]:
            self.to(self.system, reqs=self.reqs, setup_cmds=setup_cmds)

    # ----------------- Constructor helper methods -----------------

    @staticmethod
    def from_config(config: dict, dryrun: bool = True):
        """Create a Function object from a config dictionary."""
        config["reqs"] = [
            Package.from_config(package, dryrun=True)
            if isinstance(package, dict)
            else package
            for package in config["reqs"]
        ]

        if isinstance(config["system"], dict):
            config["system"] = Cluster.from_config(config["system"], dryrun=dryrun)

        if "fn_pointers" not in config:
            raise ValueError(
                "No fn_pointers provided in config. Please provide a path "
                "to a python file, module, and function name."
            )

        return Function(**config, dryrun=dryrun)

    @classmethod
    def _check_for_child_configs(cls, config):
        """Overload by child resources to load any resources they hold internally."""
        system = config["system"]
        if isinstance(system, str) and rh_config.rns_client.exists(system):
            # if the system is set to a cluster
            cluster_config: dict = rh_config.rns_client.load_config(name=system)
            if not cluster_config:
                raise Exception(f"No cluster config saved for {system}")

            # set the cluster config as the system
            config["system"] = cluster_config
        return config

    def to(
        self,
        system: Union[str, Cluster],
        reqs: Optional[List[str]] = None,
        setup_cmds: Optional[List[str]] = None,
    ):
        """
        Set up a Function on the given system, install the reqs, and run setup_cmds.

        See the args of the factory method :func:`function` for more information.
        """
        # We need to backup the system here so the __getstate__ method of the cluster
        # doesn't wipe the client and _grpc_client of this function's cluster when
        # deepcopy copies it.
        hw_backup = self.system
        self.system = None
        new_function = copy.deepcopy(self)
        self.system = hw_backup
        if isinstance(system, str):
            system = Cluster.from_name(system, dryrun=self.dryrun)
        new_function.system = system
        new_function.reqs = reqs if reqs else self.reqs
        new_function.setup_cmds = setup_cmds if setup_cmds else self.setup_cmds

        logging.info("Setting up Function on cluster.")
        new_function.system.install_packages(new_function.reqs)
        if self.setup_cmds:
            new_function.system.run(self.setup_cmds)
        logging.info("Function setup complete.")

        return new_function

    def run_setup(self, cmds: List[str], force: bool = False):
        """Run the given setup commands on the system."""
        to_run = []
        for cmd in cmds:
            if force or cmd not in self.setup_cmds:
                to_run.append(cmd)
        if to_run:
            self.setup_cmds.extend(to_run)
            self.system.run(to_run)

    @staticmethod
    def extract_fn_paths(raw_fn: Callable, reqs: List[str]):
        """Get the path to the module, module name, and function name to be able to import it on the server"""
        if not isinstance(raw_fn, Callable):
            raise TypeError(
                f"Invalid fn for Function, expected Callable but received {type(raw_fn)}"
            )
        # Background on all these dunders: https://docs.python.org/3/reference/import.html
        module = inspect.getmodule(raw_fn)

        # Need to resolve in case just filename is given
        module_path = (
            str(Path(inspect.getfile(module)).resolve())
            if hasattr(module, "__file__")
            else None
        )

        # TODO better way of detecting if in a notebook or interactive Python env
        if (
            not module_path
            or module_path.endswith("ipynb")
            or raw_fn.__name__ == "<lambda>"
        ):
            # The only time __file__ wouldn't be present is if the function is defined in an interactive
            # interpreter or a notebook. We can't import on the server in that case, so we need to cloudpickle
            # the fn to send it over. The __call__ function will serialize the function if we return it this way.
            # This is a short-term hack.
            # return None, "notebook", raw_fn.__name__
            root_path = os.getcwd()
            module_name = "notebook"
            fn_name = raw_fn.__name__
        else:
            root_path = os.path.dirname(module_path)
            module_name = inspect.getmodulename(module_path)
            # TODO __qualname__ doesn't work when fn is aliased funnily, like torch.sum
            fn_name = getattr(raw_fn, "__qualname__", raw_fn.__name__)

            # Adapted from https://github.com/modal-labs/modal-client/blob/main/modal/_function_utils.py#L94
            if getattr(module, "__package__", None):
                module_path = os.path.abspath(module.__file__)
                package_paths = [
                    os.path.abspath(p) for p in __import__(module.__package__).__path__
                ]
                base_dirs = [
                    base_dir
                    for base_dir in package_paths
                    if os.path.commonpath((base_dir, module_path)) == base_dir
                ]

                if len(base_dirs) != 1:
                    logger.info(f"Module files: {module_path}")
                    logger.info(f"Package paths: {package_paths}")
                    logger.info(f"Base dirs: {base_dirs}")
                    raise Exception("Wasn't able to find the package directory!")
                root_path = os.path.dirname(base_dirs[0])
                module_name = module.__spec__.name

        remote_import_path = None
        for req in reqs:
            local_path = None
            if not isinstance(req, str) and req.is_local():
                local_path = Path(req.local_path)
            elif isinstance(req, str):
                if req.split(":")[0] in ["local", "reqs", "pip"]:
                    req = req.split(":")[1]

                if Path(req).expanduser().resolve().exists():
                    # Relative paths are relative to the working directory in Folders/Packages!
                    local_path = (
                        Path(req).expanduser()
                        if Path(req).expanduser().is_absolute()
                        else Path(rh_config.rns_client.locate_working_dir()) / req
                    )

            if local_path:
                try:
                    # Module path relative to package
                    remote_import_path = str(
                        local_path.name / Path(root_path).relative_to(local_path)
                    )
                except ValueError:  # Not a subdirectory
                    pass
        return remote_import_path, module_name, fn_name

    # Found in python decorator logic, maybe use
    # func_name = getattr(f, '__qualname__', f.__name__)
    # module_name = getattr(f, '__module__', '')
    # if module_name:
    #     full_name = f'{module_name}.{func_name}'
    # else:
    #     full_name = func_name

    # ----------------- Function call methods -----------------

    def __call__(self, *args, stream_logs=False, **kwargs):
        fn_type = "call"
        if self.access in [ResourceAccess.WRITE, ResourceAccess.READ]:
            if not self.system or self.system.name == rh_config.obj_store.cluster_name:
                [relative_path, module_name, fn_name] = self.fn_pointers
                fn = get_fn_by_name(
                    module_name=module_name,
                    fn_name=fn_name,
                    relative_path=relative_path,
                )
                return call_fn_by_type(
                    fn, fn_type, fn_name, relative_path, args, kwargs
                )
            elif stream_logs:
                run_key = self.remote(*args, **kwargs)
                return self.system.get(run_key, stream_logs=True)
            else:
                return self._call_fn_with_ssh_access(
                    fn_type=fn_type, args=args, kwargs=kwargs
                )
        else:
            # run the function via http path - user only needs Proxy access
            if self.access != ResourceAccess.PROXY:
                raise RuntimeError("Running http path requires proxy access")
            if not rh_config.rns_client.token:
                raise ValueError(
                    "Token must be saved in the local .rh config in order to use an http path"
                )
            http_url = self.http_url()
            logger.info(f"Running {self.name} via http path: {http_url}")
            resp = requests.post(
                http_url,
                data=json.dumps({"args": args, "kwargs": kwargs}),
                headers=rh_config.rns_client.request_headers,
            )
            if resp.status_code != 200:
                raise Exception(
                    f"Failed to run Function endpoint: {load_resp_content(resp)}"
                )

            res = read_resp_data(resp)
            return res

    def repeat(self, num_repeats: int, *args, **kwargs):
        """Repeat the Function call multiple times.

        Args:
            num_repeats (int): Number of times to repeat the Function call.
            *args: Positional arguments to pass to the Function
            **kwargs: Keyword arguments to pass to the Function
        """
        if self.access in [ResourceAccess.WRITE, ResourceAccess.READ]:
            return self._call_fn_with_ssh_access(
                fn_type="repeat", args=[num_repeats, args], kwargs=kwargs
            )
        else:
            raise NotImplementedError(
                "Function.repeat only works with Write or Read access, not Proxy access"
            )

    def map(self, arg_list, **kwargs):
        """Map a function over a list of arguments."""
        if self.access in [ResourceAccess.WRITE, ResourceAccess.READ]:
            return self._call_fn_with_ssh_access(
                fn_type="map", args=arg_list, kwargs=kwargs
            )
        else:
            raise NotImplementedError(
                "Function.map only works with Write or Read access, not Proxy access"
            )

    def starmap(self, args_lists, **kwargs):
        """Like :func:`map` except that the elements of the iterable are expected to be iterables
        that are unpacked as arguments. An iterable of [(1,2), (3, 4)] results in [func(1,2), func(3,4)]."""
        if self.access in [ResourceAccess.WRITE, ResourceAccess.READ]:
            return self._call_fn_with_ssh_access(
                fn_type="starmap", args=args_lists, kwargs=kwargs
            )
        else:
            raise NotImplementedError(
                "Function.starmap only works with Write or Read access, not Proxy access"
            )

    def enqueue(self, *args, **kwargs):
        """Enqueue a Function call to be run later."""
        if self.access in [ResourceAccess.WRITE, ResourceAccess.READ]:
            return self._call_fn_with_ssh_access(
                fn_type="queue", args=args, kwargs=kwargs
            )
        else:
            raise NotImplementedError(
                "Function.enqueue only works with Write or Read access, not Proxy access"
            )

    def remote(self, *args, **kwargs):
        """Run async remote call on cluster."""
        # TODO [DG] pin the obj_ref and return a string (printed to log) so result can be retrieved later and we
        # don't need to init ray here. Also, allow user to pass the string as a param to remote().
        # TODO [DG] add rpc for listing gettaable strings, plus metadata (e.g. when it was created)
        # We need to ray init here so the returned Ray object ref doesn't throw an error it's deserialized
        # import ray
        # ray.init(ignore_reinit_error=True)
        if self.access in [ResourceAccess.WRITE, ResourceAccess.READ]:
            run_key = self._call_fn_with_ssh_access(
                fn_type="remote", args=args, kwargs=kwargs
            )
            cluster_name = (
                f'rh.cluster(name="{self.system.rns_address}")'
                if self.system.name
                else "<my_cluster>"
            )
            # TODO print this nicely
            logger.info(
                f"Submitted remote call to cluster. Result or logs can be retrieved"
                f'\n with run_key "{run_key}", e.g. '
                f'\n`{cluster_name}.get("{run_key}", stream_logs=True)` in python '
                f'\n`runhouse logs "{self.system.name}" {run_key}` from the command line.'
                f"\n or cancelled with "
                f'\n`{cluster_name}.cancel("{run_key}")` in python or '
                f'\n`runhouse cancel "{self.system.name}" {run_key}` from the command line.'
            )
            return run_key
        else:
            raise NotImplementedError(
                "Function.remote only works with Write or Read access, not Proxy access"
            )

    def get(self, obj_ref):
        """Get the result of a Function call that was submitted as async using `remote`.

        Args:
            obj_ref: A single or list of Ray.ObjectRef objects returned by a Function.remote() call. The ObjectRefs
                must be from the cluster that this Function is running on.
        """
        # TODO [DG] replace with self.system.get()?
        if self.access in [ResourceAccess.WRITE, ResourceAccess.READ]:
            arg_list = obj_ref if isinstance(obj_ref, list) else [obj_ref]
            return self._call_fn_with_ssh_access(
                fn_type="get", args=arg_list, kwargs={}
            )
        else:
            raise NotImplementedError(
                "Function.get only works with Write or Read access, not Proxy access"
            )

    def _call_fn_with_ssh_access(self, fn_type, args, kwargs):
        # https://docs.ray.io/en/latest/ray-core/tasks/patterns/map-reduce.html
        # return ray.get([map.remote(i, map_func) for i in replicas])
        # TODO allow specifying resources per worker for map
        # TODO [DG] check whether we're on the cluster and if so, just call the function directly via the
        # helper function currently in UnaryServer
        name = self.name or "anonymous function"
        if self.fn_pointers is None:
            raise RuntimeError(f"No fn pointers saved for {name}")

        [relative_path, module_name, fn_name] = self.fn_pointers
        name = self.name or fn_name or "anonymous function"
        logger.info(f"Running {name} via gRPC")
        res = self.system.run_module(
            relative_path, module_name, fn_name, fn_type, args, kwargs
        )
        return res

    # TODO [DG] test this properly
    # def debug(self, redirect_logging=False, timeout=10000, *args, **kwargs):
    #     """Run the Function in debug mode. This will run the Function through a tunnel interpreter, which
    #     allows the use of breakpoints and other debugging tools, like rh.ipython().
    #     FYI, alternative ideas from Ray: https://github.com/ray-project/ray/issues/17197
    #     FYI, alternative Modal folks shared: https://github.com/modal-labs/modal-client/pull/32
    #     """
    #     from paramiko import AutoAddPolicy
    #
    #     # Importing this here because they're heavy
    #     from plumbum.machines.paramiko_machine import ParamikoMachine
    #     from rpyc.utils.classic import redirected_stdio
    #     from rpyc.utils.zerodeploy import DeployedServer
    #
    #     creds = self.system.ssh_creds()
    #     ssh_client = ParamikoMachine(
    #         self.system.address,
    #         user=creds["ssh_user"],
    #         keyfile=str(Path(creds["ssh_private_key"]).expanduser()),
    #         missing_host_policy=AutoAddPolicy(),
    #     )
    #     server = DeployedServer(
    #         ssh_client, server_class="rpyc.utils.server.ForkingServer"
    #     )
    #     conn = server.classic_connect()
    #
    #     if redirect_logging:
    #         rlogger = conn.modules.logging.getLogger()
    #         rlogger.parent = logging.getLogger()
    #
    #     conn._config[
    #         "sync_request_timeout"
    #     ] = timeout  # seconds. May need to be longer for real debugging.
    #     conn._config["allow_public_attrs"] = True
    #     conn._config["allow_pickle"] = True
    #     # This assumes the code is already synced over to the remote container
    #     remote_fn = getattr(conn.modules[self.fn.__module__], self.fn.__name__)
    #
    #     with redirected_stdio(conn):
    #         res = remote_fn(*args, **kwargs)
    #     conn.close()
    #     return res

    @property
    def config_for_rns(self):
        # TODO save Package resource, because fn_pointers are meaningless without the package.

        config = super().config_for_rns

        config.update(
            {
                "system": self._resource_string_for_subconfig(self.system),
                "reqs": [
                    self._resource_string_for_subconfig(package)
                    for package in self.reqs
                ],
                "setup_cmds": self.setup_cmds,
                "fn_pointers": self.fn_pointers,
            }
        )
        return config

    def _save_sub_resources(self):
        self.system.save()

    # TODO maybe reuse these if we starting putting each function in its own container
    # @staticmethod
    # def run_ssh_cmd_in_cluster(ssh_key, ssh_user, address, cmd, port_fwd=None):
    #     subprocess.run("ssh -tt -o IdentitiesOnly=yes -i "
    #                    f"{ssh_key} {port_fwd or ''}"
    #                    f"{ssh_user}@{address} docker exec -it ray_container /bin/bash -c {cmd}".split(' '))

    def ssh(self):
        """SSH into the system."""
        if self.system is None:
            raise RuntimeError("System must be specified and up to ssh into a Function")
        self.system.ssh()

    def send_secrets(self, providers: Optional[List[str]] = None):
        """Send secrets to the system."""
        self.system.send_secrets(providers=providers)

    def http_url(self, curl_command=False, *args, **kwargs) -> str:
        """
        Return the endpoint needed to run the Function on the remote cluster, or provide the curl command if requested.
        """
        raise NotImplementedError("http_url not yet implemented for Function")
        resource_uri = rh_config.rns_client.resource_uri(name=self.name)
        uri = f"proxy/{resource_uri}"
        if curl_command:
            # NOTE: curl command should include args and kwargs - this will help us generate better API docs
            if not is_jsonable(args) or not is_jsonable(kwargs):
                raise Exception(
                    "Invalid Function func params provided, must be able to convert args and kwargs to json"
                )

            return (
                "curl -X 'POST' '{api_server_url}/proxy{resource_uri}/endpoint' "
                "-H 'accept: application/json' "
                "-H 'Authorization: Bearer {auth_token}' "
                "-H 'Content-Type: application/json' "
                "-d '{data}'".format(
                    api_server_url=rh_config.rns_client.api_server_url,
                    resource_uri=uri,
                    auth_token=rh_config.rns_client.token,
                    data=json.dumps({"args": args, "kwargs": kwargs}),
                )
            )

        # HTTP URL needed to run the Function remotely
        http_url = f"{rh_config.rns_client.api_server_url}/{uri}/endpoint"
        return http_url

    def notebook(self, persist=False, sync_package_on_close=None, port_forward=8888):
        """Tunnel into and launch notebook from the system."""
        # Roughly trying to follow:
        # https://towardsdatascience.com/using-jupyter-notebook-running-on-a-remote-docker-container-via-ssh-ea2c3ebb9055
        # https://docs.ray.io/en/latest/ray-core/using-ray-with-jupyter.html
        if self.system is None:
            raise RuntimeError("Cannot SSH, running locally")

        tunnel, port_fwd = self.system.ssh_tunnel(
            local_port=port_forward, num_ports_to_try=10
        )
        try:
            install_cmd = "pip install jupyterlab"
            jupyter_cmd = f"jupyter lab --port {port_fwd} --no-browser"
            # port_fwd = '-L localhost:8888:localhost:8888 '  # TOOD may need when we add docker support
            with self.system.pause_autostop():
                self.system.run(commands=[install_cmd, jupyter_cmd], stream_logs=True)

        finally:
            if sync_package_on_close:
                if sync_package_on_close == "./":
                    sync_package_on_close = rh_config.rns_client.locate_working_dir()
                pkg = Package.from_string("local:" + sync_package_on_close)
                self.system.rsync(source=f"~/{pkg.name}", dest=pkg.local_path, up=False)
            if not persist:
                tunnel.stop()
                kill_jupyter_cmd = f"jupyter notebook stop {port_fwd}"
                self.system.run(commands=[kill_jupyter_cmd])

    def keep_warm(
        self,
        autostop_mins=None,
        # TODO regions: List[str] = None,
        # TODO min_replicas: List[int] = None,
        # TODO max_replicas: List[int] = None
    ):
        """Keep the system warm for autostop_mins. If autostop_mins is ``None`` or -1, keep warm indefinitely."""
        if autostop_mins is None:
            logger.info(f"Keeping {self.name} indefinitely warm")
            # keep indefinitely warm if user doesn't specify
            autostop_mins = -1
        self.system.keep_warm(autostop_mins=autostop_mins)

    @staticmethod
    def _handle_nb_fn(fn, fn_pointers, serialize_notebook_fn, name):
        """Handle the case where the user passes in a notebook function"""
        if serialize_notebook_fn:
            # This will all be cloudpickled by the gRPC client and unpickled by the gRPC server
            # Note that this means the function cannot be saved, and it's better that way because
            # pickling functions is not meant for long term storage. Case in point, this method will be
            # sensitive to differences in minor Python versions between the serializing and deserializing envs.
            return "", "notebook", fn
        else:
            # TODO put this in the current folder instead?
            module_path = Path.cwd() / (f"{name}_fn.py" if name else "sent_fn.py")
            logging.info(
                f"Writing out function function to {str(module_path)}. Please make "
                f"sure the function does not rely on any local variables, "
                f"including imports (which should be moved inside the function body)."
            )
            if not name:
                logging.warning(
                    "You should name Functions that are created in notebooks to avoid naming collisions "
                    "between the modules that are created to hold their functions "
                    '(i.e. "sent_fn.py" errors.'
                )
            source = inspect.getsource(fn).strip()
            with module_path.open("w") as f:
                f.write(source)
            return fn_pointers[0], module_path.stem, fn_pointers[2]
            # from importlib.util import spec_from_file_location, module_from_spec
            # spec = spec_from_file_location(config['name'], str(module_path))
            # module = module_from_spec(spec)
            # spec.loader.exec_module(module)
            # new_fn = getattr(module, fn_pointers[2])
            # fn_pointers = Function.extract_fn_paths(raw_fn=new_fn, reqs=config['reqs'])


def function(
    fn: Optional[Union[str, Callable]] = None,
    name: Optional[str] = None,
    system: Optional[Union[str, Cluster]] = None,
    reqs: Optional[List[str]] = None,
    setup_cmds: Optional[List[str]] = None,
    # TODO image: Optional[str] = None,
    dryrun: bool = False,
    load_secrets: bool = False,
    serialize_notebook_fn: bool = False,
    load: bool = True,
):
    """Factory method for constructing a Runhouse Function object.

    Args:
        fn (Optional[str or Callable]): The function to execute on the remote system when the function is called.
        name (Optional[str]): Name of the Function to create or retrieve.
            This can be either from a local config or from the RNS.
        system (Optional[str or Cluster]): Hardware (cluster) on which to execute the Function.
            This can be either the string name of a Cluster object, or a Cluster object.
        reqs (Optional[List[str]]): List of requirements to install on the remote cluster, or path to the
            requirements.txt file.
        dryrun (bool): Whether to create the Function if it doesn't exist, or load the Function object as a dryrun.
            (Default: ``False``)
        load_secrets (bool): Whether or not to send secrets; only applicable if `dryrun` is set to ``False``.
            (Default: ``False``)
        serialize_notebook_fn (bool): If function is of a notebook setting, whether or not to serialized the function.
            (Default: ``False``)
        load (bool): Whether to load an existing config for the Function. (Default: ``True``)

    Returns:
        Function: The resulting Function object.

    Example:
        >>> def sum(a, b):
        >>>    return a + b
        >>>
        >>> # creating the function
        >>> summer = rh.function(fn=sum, system=cluster, reqs=['requirements.txt'])
        >>> # or, equivalently
        >>> summer = rh.function(fn=sum).to(cluster, reqs=['requirements.txt'])
        >>>
        >>> # using the function
        >>> summer(5, 8)  # returns 13
    """

    config = rh_config.rns_client.load_config(name) if load else {}
    config["name"] = name or config.get("rns_address", None) or config.get("name")
    config["reqs"] = reqs if reqs is not None else config.get("reqs", [])

    processed_reqs = []
    for req in config["reqs"]:
        # TODO [DG] the following is wrong. RNS address doesn't have to start with '/'. However if we check if each
        #  string exists in RNS this will be incredibly slow, so leave it for now.
        if isinstance(req, str) and req[0] == "/" and rh_config.rns_client.exists(req):
            # If req is an rns address
            req = rh_config.rns_client.load_config(req)
        processed_reqs.append(req)
    config["reqs"] = processed_reqs

    if callable(fn):
        if not [req for req in config["reqs"] if "./" in req]:
            config["reqs"].append("./")
        fn_pointers = Function.extract_fn_paths(raw_fn=fn, reqs=config["reqs"])
        if fn_pointers[1] == "notebook":
            fn_pointers = Function._handle_nb_fn(
                fn,
                fn_pointers=fn_pointers,
                serialize_notebook_fn=serialize_notebook_fn,
                name=fn_pointers[2] or config["name"],
            )
        config["fn_pointers"] = fn_pointers
    elif isinstance(fn, str):
        # Url must match a regex of the form
        # 'https://github.com/username/repo_name/blob/branch_name/path/to/file.py:func_name'
        # Use a regex to extract username, repo_name, branch_name, path/to/file.py, and func_name
        pattern = (
            r"https://github\.com/(?P<username>[^/]+)/(?P<repo_name>[^/]+)/blob/"
            r"(?P<branch_name>[^/]+)/(?P<path>[^:]+):(?P<func_name>.+)"
        )
        match = re.match(pattern, fn)

        if match:
            username = match.group("username")
            repo_name = match.group("repo_name")
            branch_name = match.group("branch_name")
            path = match.group("path")
            func_name = match.group("func_name")
        else:
            raise ValueError(
                "fn must be a callable or string of the form "
                '"https://github.com/username/repo_name/blob/branch_name/path/to/file.py:func_name"'
            )
        module_name = Path(path).stem
        relative_path = str(repo_name / Path(path).parent)
        config["fn_pointers"] = (relative_path, module_name, func_name)
        # TODO [DG] check if the user already added this in their reqs
        repo_package = git_package(
            git_url=f"https://github.com/{username}/{repo_name}.git",
            revision=branch_name,
        )
        config["reqs"].insert(0, repo_package)

    config["system"] = system or config.get("system")
    if isinstance(config["system"], str):
        hw_dict = rh_config.rns_client.load_config(config["system"])
        if not hw_dict:
            raise RuntimeError(
                f'Cluster {rh_config.rns_client.resolve_rns_path(config["system"])} '
                f"not found locally or in RNS."
            )
        config["system"] = hw_dict

    config["setup_cmds"] = (
        setup_cmds if setup_cmds is not None else config.get("setup_cmds")
    )

    config["access_level"] = config.get("access_level", Function.DEFAULT_ACCESS)

    new_function = Function.from_config(config, dryrun=dryrun)

    if load_secrets and not dryrun:
        new_function.send_secrets()

    return new_function



from typing import List, Optional, Union

from runhouse.rh_config import rns_client

from .cluster import Cluster
from .on_demand_cluster import OnDemandCluster


# Cluster factory method
def cluster(
    name: str,
    instance_type: Optional[str] = None,
    num_instances: Optional[int] = None,
    provider: Optional[str] = None,
    autostop_mins: Optional[int] = None,
    use_spot: bool = False,
    image_id: Optional[str] = None,
    region: Optional[str] = None,
    ips: List[str] = None,
    ssh_creds: Optional[dict] = None,
    dryrun: bool = False,
    load: bool = True,
) -> Union[Cluster, OnDemandCluster]:
    """
    Builds an instance of :class:`Cluster`.

    Args:
        name (str): Name for the cluster, to re-use later on.
        instance_type (int, optional): Type of cloud instance to use for the cluster. This could
            be a Runhouse built-in type, or your choice of instance type.
        num_instances (int, optional): Number of instances to use for the cluster.
        provider (str, optional): Cloud provider to use for the cluster.
        autostop_mins (int, optional): Number of minutes to keep the cluster up after inactivity,
            or ``-1`` to keep cluster up indefinitely.
        use_spot (bool, optional): Whether or not to use spot instance.
        image_id (str, optional): Custom image ID for the cluster.
        region (str, optional): The region to use for the cluster.
        ips (List[str], optional): List of IP addresses for the BYO cluster.
        ssh_creds (dict, optional): Dictionary mapping SSH credentials.
            Example: ``ssh_creds={'ssh_user': '...', 'ssh_private_key':'<path_to_key>'}``
        dryrun (bool): Whether to create the Cluster if it doesn't exist, or load a Cluster object as a dryrun.
            (Default: ``False``)
        load (bool): Whether to load an existing config for the Cluster. (Default: ``True``)

    Returns:
        Cluster or OnDemandCluster: The resulting cluster.

    Example:
        >>> # BYO Cluster
        >>> gpu = rh.cluster(ips=['<ip of the cluster>'],
        >>>          ssh_creds={'ssh_user': '...', 'ssh_private_key':'<path_to_key>'},
        >>>          name='rh-a10x')

        >>> # On-Demand SkyPilot Cluster (OnDemandCluster)
        >>> gpu = rh.cluster(name='rh-4-a100s',
        >>>                  instance_type='A100:4',
        >>>                  provider='gcp',
        >>>                  autostop_mins=-1,
        >>>                  use_spot=True,
        >>>                  image_id='my_ami_string',
        >>>                  region='us-east-1',
        >>>                  )
    """
    config = rns_client.load_config(name) if load else {}
    config["name"] = name or config.get("rns_address", None) or config.get("name")
    config["ips"] = ips or config.get("ips", None)
    # ssh creds should only be in Secrets management, not in config
    config["ssh_creds"] = ssh_creds or config.get("ssh_creds", None)
    if config["ips"]:
        return Cluster.from_config(config, dryrun=dryrun)

    config["instance_type"] = instance_type or config.get("instance_type", None)
    config["num_instances"] = num_instances or config.get("num_instances", None)
    config["provider"] = provider or config.get("provider", None)
    config["autostop_mins"] = (
        autostop_mins
        if autostop_mins is not None
        else config.get("autostop_mins", None)
    )
    config["use_spot"] = (
        use_spot if use_spot is not None else config.get("use_spot", None)
    )
    config["image_id"] = (
        image_id if image_id is not None else config.get("image_id", None)
    )
    config["region"] = region if region is not None else config.get("region", None)

    new_cluster = OnDemandCluster.from_config(config, dryrun=dryrun)

    return new_cluster



import logging

from .. import OnDemandCluster
from ..api_utils.utils import generate_uuid

from .table import Table

logger = logging.getLogger(__name__)


class PandasTable(Table):
    DEFAULT_FOLDER_PATH = "/runhouse/pandas-tables"
    DEFAULT_STREAM_FORMAT = "pandas"

    def __init__(self, **kwargs):
        if not kwargs.get("file_name"):
            kwargs["file_name"] = f"{generate_uuid()}.parquet"
        super().__init__(**kwargs)

    def __iter__(self):
        for block in self.stream(batch_size=self.DEFAULT_BATCH_SIZE):
            for idx, row in block.iterrows():
                yield row

    @staticmethod
    def from_config(config: dict, dryrun=True):
        """Load config values into the object."""
        if isinstance(config["system"], dict):
            config["system"] = OnDemandCluster.from_config(
                config["system"], dryrun=dryrun
            )
        return PandasTable(**config, dryrun=dryrun)

    def write(self):
        if self._cached_data is not None:
            # https://pandas.pydata.org/pandas-docs/version/1.1/reference/api/pandas.DataFrame.to_parquet.html
            self.data.to_parquet(
                self.fsspec_url,
                storage_options=self.data_config,
                partition_cols=self.partition_cols,
            )

        return self

    def fetch(self, **kwargs):
        import pandas as pd

        # https://pandas.pydata.org/docs/reference/api/pandas.read_parquet.html
        self._cached_data = pd.read_parquet(
            self.fsspec_url, storage_options=self.data_config
        )
        return self._cached_data



import copy
import logging
from pathlib import Path
from typing import Dict, Optional

import runhouse as rh
from runhouse.rh_config import rns_client
from runhouse.rns.api_utils.utils import generate_uuid
from runhouse.rns.folders.folder import Folder, folder
from runhouse.rns.obj_store import _current_cluster
from runhouse.rns.resource import Resource

logger = logging.getLogger(__name__)


class Blob(Resource):
    RESOURCE_TYPE = "blob"
    DEFAULT_FOLDER_PATH = "/runhouse-blob"
    DEFAULT_CACHE_FOLDER = ".cache/runhouse/blobs"

    def __init__(
        self,
        path: Optional[str] = None,
        name: Optional[str] = None,
        system: Optional[str] = Folder.DEFAULT_FS,
        data_config: Optional[Dict] = None,
        dryrun: bool = False,
        **kwargs,
    ):
        """
        Runhouse Blob object

        .. note::
                To build a Blob, please use the factory method :func:`blob`.
        """
        super().__init__(name=name, dryrun=dryrun)
        self._filename = str(Path(path).name) if path else self.name
        # Use factory method so correct subclass for system is returned
        self._folder = folder(
            path=str(Path(path).parent) if path is not None else path,
            system=system,
            data_config=data_config,
            dryrun=dryrun,
        )
        self._cached_data = None

    @property
    def config_for_rns(self):
        config = super().config_for_rns
        blob_config = {
            "path": self.path,  # pair with data source to create the physical URL
            "resource_type": self.RESOURCE_TYPE,
            "system": self.system,
        }
        config.update(blob_config)
        return config

    @staticmethod
    def from_config(config: dict, dryrun=True):
        return Blob(**config, dryrun=dryrun)

    @property
    def data(self):
        """Get the blob data."""
        if self._cached_data is not None:
            return self._cached_data
        data = self.fetch()
        return data

    @data.setter
    def data(self, new_data):
        """Update the data blob to new data."""
        self._cached_data = new_data

    @property
    def system(self):
        return self._folder.system

    @system.setter
    def system(self, new_system):
        self._folder.system = new_system

    @property
    def path(self):
        return self._folder.path + "/" + self._filename

    @path.setter
    def path(self, new_path):
        self._folder.path = str(Path(new_path).parent)
        self._filename = str(Path(new_path).name)

    @property
    def data_config(self):
        return self._folder.data_config

    @data_config.setter
    def data_config(self, new_data_config):
        self._folder.data_config = new_data_config

    @property
    def fsspec_url(self):
        return self._folder.fsspec_url + "/" + self._filename

    def open(self, mode: str = "rb"):
        """Get a file-like (OpenFile container object) of the blob data. User must close the file, or use this
        method inside of a with statement (e.g. `with my_blob.open() as f:`)."""
        return self._folder.open(self._filename, mode=mode)

    def to(
        self, system, path: Optional[str] = None, data_config: Optional[dict] = None
    ):
        """Return a copy of the blob on the destination system and path."""
        new_blob = copy.copy(self)
        new_blob._folder = self._folder.to(
            system=system, path=path, data_config=data_config
        )
        return new_blob

    def fetch(self):
        """Return the data for the user to deserialize"""
        self._cached_data = self._folder.get(self._filename)
        return self._cached_data

    def _save_sub_resources(self):
        if isinstance(self.system, Resource):
            self.system.save()

    def write(self):
        """Save the underlying blob to its specified fsspec URL."""
        # TODO figure out default behavior for not overwriting but still saving
        # if not overwrite:
        #     TODO check if data_url is already in use
        #     time = datetime.today().strftime('%Y-%m-%d_%H:%M:%S')
        #     self.data_url = self.data_url + time or time

        # TODO check if self._cached_data is None, and if so, don't just download it to then save it again?
        with self.open(mode="wb") as f:
            if not isinstance(self.data, bytes):
                # Avoid TypeError: a bytes-like object is required
                raise TypeError(
                    f"Cannot save blob with data of type {type(self.data)}, data must be serialized"
                )

            f.write(self.data)

        return self

    def delete_in_system(self):
        """Delete the blob and the folder it lives in from the file system."""
        self._folder.rm(self._filename)
        if self.system == "file":
            # Deleting the blob itself in a local file system will not remove the parent folder by default
            self._folder.delete_in_system()

    def exists_in_system(self):
        """Check whether the blob exists in the file system"""
        return self._folder.fsspec_fs.exists(self.fsspec_url)

    # TODO [DG] get rid of this in favor of just "sync_down(path, system)" ?
    def sync_from_cluster(self, cluster, path: Optional[str] = None):
        """Efficiently rsync down a blob from a cluster, into the path of the current Blob object."""
        if not cluster.address:
            raise ValueError("Cluster must be started before copying data to it.")

        cluster.rsync(source=self.path, dest=path, up=False)


def blob(
    data=None,
    name: Optional[str] = None,
    path: Optional[str] = None,
    system: Optional[str] = None,
    data_config: Optional[Dict] = None,
    mkdir: bool = False,
    dryrun: bool = False,
    load: bool = True,
):
    """Returns a Blob object, which can be used to interact with the resource at the given path

    Args:
        data: Blob data. This should be provided as a serialized object.
        name (Optional[str]): Name to give the blob object, to be reused later on.
        path (Optional[str]): Path (or path) of the blob object.
        system (Optional[str]): File system. Currently this must be one of:
           [``file``, ``github``, ``sftp``, ``ssh``,``s3``, ``gs``, ``azure``].
            We are working to add additional file system support.
        data_config (Optional[Dict]): The data config to pass to the underlying fsspec handler.
        mkdir (bool): Whether to create a remote folder for the blob. (Default: ``False``)
        dryrun (bool): Whether to create the Blob if it doesn't exist, or load a Blob object as a dryrun.
            (Default: ``False``)
        load (bool): Whether to load an existing config for the Blob. (Default: ``True``)

    Returns:
        Blob: The resulting blob.

    Example:
        >>> data = json.dumps(list(range(50))
        >>>
        >>> # Remote blob with name and no path (saved to bucket called runhouse/blobs/my-blob)
        >>> rh.blob(name="@/my-blob", data=data, data_source='s3', dryrun=False)
        >>>
        >>> # Remote blob with name and path
        >>> rh.blob(name='@/my-blob', path='/runhouse-tests/my_blob.pickle', data=data, system='s3', dryrun=False)
        >>>
        >>> # Local blob with name and path, save to local filesystem
        >>> rh.blob(name=name, data=data, path=str(Path.cwd() / "my_blob.pickle"), dryrun=False)
        >>>
        >>> # Local blob with name and no path (saved to ~/.cache/blobs/my-blob)
        >>> rh.blob(name="~/my-blob", data=data, dryrun=False)

        >>> # Loading a blob
        >>> my_local_blob = rh.blob(name="~/my_blob")
        >>> my_s3_blob = rh.blob(name="@/my_blob")
    """
    config = rns_client.load_config(name) if load else {}

    system = (
        system
        or config.get("system")
        or _current_cluster(key="config")
        or Folder.DEFAULT_FS
    )
    config["system"] = system

    name = name or config.get("rns_address") or config.get("name")

    data_path = path or config.get("path")
    if data_path is None:
        blob_name_in_path = (
            f"{generate_uuid()}/{rns_client.resolve_rns_data_resource_name(name)}"
        )

        if (
            system == rns_client.DEFAULT_FS
            or config["system"] == _current_cluster()
            or (
                isinstance(config["system"], dict)
                and config["system"]["name"] == _current_cluster()
            )
        ):
            # create random path to store in .cache folder of local filesystem
            data_path = str(
                Path(f"~/{Blob.DEFAULT_CACHE_FOLDER}/{blob_name_in_path}").expanduser()
            )
        else:
            # save to the default bucket
            data_path = f"{Blob.DEFAULT_FOLDER_PATH}/{blob_name_in_path}"

    config["name"] = name
    config["path"] = data_path
    config["data_config"] = data_config or config.get("data_config")

    if isinstance(config["system"], str) and rns_client.exists(
        config["system"], resource_type="cluster"
    ):
        config["system"] = rns_client.load_config(config["system"])
    elif isinstance(config["system"], dict):
        from runhouse.rns.hardware.cluster import Cluster

        config["system"] = Cluster.from_config(config["system"])

    if mkdir:
        # create the remote folder for the blob
        folder_path = str(Path(data_path).parent)
        rh.folder(name=folder_path, system=system, dryrun=True).mkdir()

    new_blob = Blob.from_config(config, dryrun=dryrun)
    new_blob.data = data

    return new_blob



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import runhouse.servers.grpc.unary_pb2 as unary__pb2


class UnaryStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.RunModule = channel.unary_unary(
            "/unary.Unary/RunModule",
            request_serializer=unary__pb2.Message.SerializeToString,
            response_deserializer=unary__pb2.MessageResponse.FromString,
        )
        self.InstallPackages = channel.unary_unary(
            "/unary.Unary/InstallPackages",
            request_serializer=unary__pb2.Message.SerializeToString,
            response_deserializer=unary__pb2.MessageResponse.FromString,
        )
        self.ClearPins = channel.unary_unary(
            "/unary.Unary/ClearPins",
            request_serializer=unary__pb2.Message.SerializeToString,
            response_deserializer=unary__pb2.MessageResponse.FromString,
        )
        self.CancelRun = channel.unary_unary(
            "/unary.Unary/CancelRun",
            request_serializer=unary__pb2.Message.SerializeToString,
            response_deserializer=unary__pb2.MessageResponse.FromString,
        )
        self.PutObject = channel.unary_unary(
            "/unary.Unary/PutObject",
            request_serializer=unary__pb2.Message.SerializeToString,
            response_deserializer=unary__pb2.MessageResponse.FromString,
        )
        self.AddSecrets = channel.unary_unary(
            "/unary.Unary/AddSecrets",
            request_serializer=unary__pb2.Message.SerializeToString,
            response_deserializer=unary__pb2.MessageResponse.FromString,
        )
        self.GetObject = channel.unary_stream(
            "/unary.Unary/GetObject",
            request_serializer=unary__pb2.Message.SerializeToString,
            response_deserializer=unary__pb2.MessageResponse.FromString,
        )


class UnaryServicer(object):
    """Missing associated documentation comment in .proto file."""

    def RunModule(self, request, context):
        """A simple RPC.

        Obtains the MessageResponse at a given position.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def InstallPackages(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ClearPins(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def CancelRun(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def PutObject(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def AddSecrets(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetObject(self, request, context):
        """streaming RPC"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_UnaryServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "RunModule": grpc.unary_unary_rpc_method_handler(
            servicer.RunModule,
            request_deserializer=unary__pb2.Message.FromString,
            response_serializer=unary__pb2.MessageResponse.SerializeToString,
        ),
        "InstallPackages": grpc.unary_unary_rpc_method_handler(
            servicer.InstallPackages,
            request_deserializer=unary__pb2.Message.FromString,
            response_serializer=unary__pb2.MessageResponse.SerializeToString,
        ),
        "ClearPins": grpc.unary_unary_rpc_method_handler(
            servicer.ClearPins,
            request_deserializer=unary__pb2.Message.FromString,
            response_serializer=unary__pb2.MessageResponse.SerializeToString,
        ),
        "CancelRun": grpc.unary_unary_rpc_method_handler(
            servicer.CancelRun,
            request_deserializer=unary__pb2.Message.FromString,
            response_serializer=unary__pb2.MessageResponse.SerializeToString,
        ),
        "PutObject": grpc.unary_unary_rpc_method_handler(
            servicer.PutObject,
            request_deserializer=unary__pb2.Message.FromString,
            response_serializer=unary__pb2.MessageResponse.SerializeToString,
        ),
        "AddSecrets": grpc.unary_unary_rpc_method_handler(
            servicer.AddSecrets,
            request_deserializer=unary__pb2.Message.FromString,
            response_serializer=unary__pb2.MessageResponse.SerializeToString,
        ),
        "GetObject": grpc.unary_stream_rpc_method_handler(
            servicer.GetObject,
            request_deserializer=unary__pb2.Message.FromString,
            response_serializer=unary__pb2.MessageResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "unary.Unary", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class Unary(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def RunModule(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/unary.Unary/RunModule",
            unary__pb2.Message.SerializeToString,
            unary__pb2.MessageResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def InstallPackages(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/unary.Unary/InstallPackages",
            unary__pb2.Message.SerializeToString,
            unary__pb2.MessageResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ClearPins(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/unary.Unary/ClearPins",
            unary__pb2.Message.SerializeToString,
            unary__pb2.MessageResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def CancelRun(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/unary.Unary/CancelRun",
            unary__pb2.Message.SerializeToString,
            unary__pb2.MessageResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def PutObject(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/unary.Unary/PutObject",
            unary__pb2.Message.SerializeToString,
            unary__pb2.MessageResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def AddSecrets(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/unary.Unary/AddSecrets",
            unary__pb2.Message.SerializeToString,
            unary__pb2.MessageResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetObject(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_stream(
            request,
            target,
            "/unary.Unary/GetObject",
            unary__pb2.Message.SerializeToString,
            unary__pb2.MessageResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



import os
from pathlib import Path
from typing import Optional

from runhouse import Secrets


class GitHubSecrets(Secrets):
    PROVIDER_NAME = "github"
    CREDENTIALS_FILE = os.path.expanduser("~/.config/gh/hosts.yml")

    @classmethod
    def read_secrets(cls, from_env: bool = False, file_path: Optional[str] = None):
        if from_env:
            raise NotImplementedError(
                f"Reading secrets from env is not supported for {cls.PROVIDER_NAME}"
            )
        else:
            creds_file = file_path or cls.default_credentials_path()
            config_data = cls.read_yaml_file(creds_file)
            token = config_data["github.com"]["oauth_token"]

        return {"token": token}

    @classmethod
    def save_secrets(cls, secrets: dict, overwrite: bool = False):

        dest_path = cls.default_credentials_path()
        cls.check_secrets_for_mismatches(
            secrets_to_save=secrets, secrets_path=dest_path, overwrite=overwrite
        )

        config = cls.read_yaml_file(dest_path) if cls.file_exists(dest_path) else {}
        config["github.com"] = {"oauth_token": secrets["token"]}

        Path(dest_path).parent.mkdir(parents=True, exist_ok=True)

        cls.save_to_yaml_file(config, dest_path)
        cls.add_provider_to_rh_config()



import os
import sys
from typing import Optional

from runhouse import Secrets


class GCPSecrets(Secrets):
    PROVIDER_NAME = "gcp"
    CREDENTIALS_FILE = os.path.expanduser(
        "~/.config/gcloud/application_default_credentials.json"
    )

    @classmethod
    def read_secrets(cls, from_env: bool = False, file_path: Optional[str] = None):
        if from_env:
            client_id = os.getenv("CLIENT_ID")
            client_secret = os.getenv("CLIENT_SECRET")
            if not client_id or not client_secret:
                raise Exception(
                    f"CLIENT_ID and CLIENT_SECRET must be set for {cls.PROVIDER_NAME}"
                )
        else:
            creds_file = file_path or cls.default_credentials_path()
            config_data = cls.read_json_file(creds_file)
            client_id = config_data["client_id"]
            client_secret = config_data["client_secret"]

        return {
            "client_id": client_id,
            "client_secret": client_secret,
        }

    @classmethod
    def save_secrets(cls, secrets: dict, overwrite: bool = False):

        dest_path = cls.default_credentials_path()
        cls.check_secrets_for_mismatches(
            secrets_to_save=secrets, secrets_path=dest_path, overwrite=overwrite
        )

        config = cls.read_json_file(dest_path) if cls.file_exists(dest_path) else {}
        config["client_id"] = secrets["client_id"]
        config["client_secret"] = secrets["client_secret"]

        # We need to do extra stuff if we're in a colab
        if "google.colab" in sys.modules:
            from rich.console import Console

            console = Console()
            console.print(
                "Please do the following to complete gcp secrets setup:",
                style="bold yellow",
            )
            console.print("!gcloud init", style="bold yellow")
            console.print("!gcloud auth application-default login", style="bold yellow")
            console.print(
                "!cp -r /content/.config/* ~/.config/gcloud", style="bold yellow"
            )

        cls.save_to_json_file(config, dest_path)
        cls.add_provider_to_rh_config()



import logging

from .. import OnDemandCluster

from .table import Table

logger = logging.getLogger(__name__)


class RayTable(Table):
    DEFAULT_FOLDER_PATH = "/runhouse/ray-tables"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @staticmethod
    def from_config(config: dict, dryrun=True):
        """Load config values into the object."""
        if isinstance(config["system"], dict):
            config["system"] = OnDemandCluster.from_config(
                config["system"], dryrun=dryrun
            )
        return RayTable(**config, dryrun=dryrun)

    def write(self):
        if self._cached_data is not None:
            self.write_ray_dataset(self.data)
            logger.info(f"Saved {str(self)} to: {self.fsspec_url}")

        return self

    def fetch(self, **kwargs):
        import ray

        self._cached_data = ray.data.read_parquet(
            self.fsspec_url, filesystem=self._folder.fsspec_fs
        )
        return self._cached_data



import os
import subprocess
from pathlib import Path
from typing import Optional

from runhouse import Secrets


class HuggingFaceSecrets(Secrets):
    PROVIDER_NAME = "huggingface"
    CREDENTIALS_FILE = os.path.expanduser("~/.huggingface/token")

    @classmethod
    def read_secrets(cls, from_env: bool = False, file_path: Optional[str] = None):
        if from_env:
            raise NotImplementedError(
                f"Reading secrets from env is not supported for {cls.PROVIDER_NAME}"
            )
        else:
            creds_path = file_path or cls.default_credentials_path()
            token = Path(creds_path).read_text()

        return {"token": token}

    @classmethod
    def save_secrets(cls, secrets: dict, overwrite: bool = False):
        # TODO check properly if hf needs to be installed
        try:
            import huggingface_hub
        except ModuleNotFoundError:
            subprocess.run(["pip", "install", "--upgrade", "huggingface-hub"])
            import huggingface_hub

        dest_path = cls.default_credentials_path()
        cls.check_secrets_for_mismatches(
            secrets_to_save=secrets, secrets_path=dest_path, overwrite=overwrite
        )

        huggingface_hub.login(token=secrets["token"])
        cls.add_provider_to_rh_config()



import logging
import re
import sys
import time

import grpc

import ray.cloudpickle as pickle

import runhouse.servers.grpc.unary_pb2 as pb2

import runhouse.servers.grpc.unary_pb2_grpc as pb2_grpc

logger = logging.getLogger(__name__)


class OutputType:
    STDOUT = "stdout"
    STDERR = "stderr"
    RESULT = "result"


class UnaryClient(object):
    """
    Client for gRPC functionality
    # TODO rename ClusterClient
    """

    DEFAULT_PORT = 50052
    MAX_MESSAGE_LENGTH = 1 * 1024 * 1024 * 1024  # 1 GB
    TIMEOUT_SEC = 3

    def __init__(self, host, port=DEFAULT_PORT):
        self.host = host
        self.port = port
        self.channel = grpc.insecure_channel(
            f"{self.host}:{self.port}",
            options=[
                ("grpc.max_send_message_length", self.MAX_MESSAGE_LENGTH),
                ("grpc.max_receive_message_length", self.MAX_MESSAGE_LENGTH),
            ],
        )
        self._connectivity_state = None

        def on_state_change(state):
            self._connectivity_state = state

        self.channel.subscribe(on_state_change, False)

        # bind the client and the server
        self.stub = pb2_grpc.UnaryStub(self.channel)

        # NOTE: might be helpful for debugging purposes
        # os.environ['GRPC_TRACE'] = 'all'
        # os.environ['GRPC_VERBOSITY'] = 'DEBUG'

    def install_packages(self, to_install):
        message = pb2.Message(message=pickle.dumps(to_install))
        server_res = self.stub.InstallPackages(message)
        [res, fn_exception, fn_traceback] = pickle.loads(server_res.message)
        if fn_exception is not None:
            logger.error(f"Error installing packages {to_install}: {fn_exception}")
            logger.error(f"Traceback: {fn_traceback}")
            raise fn_exception
        return res

    def add_secrets(self, secrets):
        message = pb2.Message(message=secrets)
        server_res = self.stub.AddSecrets(message)
        return pickle.loads(server_res.message)

    def cancel_runs(self, keys, force=False):
        message = pb2.Message(message=pickle.dumps((keys, force)))
        res = self.stub.CancelRun(message)
        return pickle.loads(res.message)

    # TODO [DG]: maybe just merge cancel into this so we can get log streaming back as we cancel a job
    def get_object(self, key, stream_logs=False):
        """
        Get a value from the server
        """
        message = pb2.Message(message=pickle.dumps((key, stream_logs)))
        for resp in self.stub.GetObject(message):
            output_type = resp.output_type
            server_res = pickle.loads(resp.message)
            if output_type == OutputType.STDOUT:
                # Regex to match tqdm progress bars
                tqdm_regex = re.compile(r"(.+)%\|(.+)\|\s+(.+)/(.+)")
                for line in server_res:
                    if tqdm_regex.match(line):
                        # tqdm lines are always preceded by a \n, so we can use \x1b[1A to move the cursor up one line
                        # For some reason, doesn't work in PyCharm's console, but works in the terminal
                        print("\x1b[1A\r" + line, end="", flush=True)
                    else:
                        print(line, end="", flush=True)
            elif output_type == OutputType.STDERR:
                for line in server_res:
                    print(line, file=sys.stderr)
            else:
                [res, fn_exception, fn_traceback] = server_res
                if fn_exception is not None:
                    logger.error(
                        f"Error running or getting run_key {key}: {fn_exception}."
                    )
                    logger.error(f"Traceback: {fn_traceback}")
                    raise fn_exception
                return res

    def put_object(self, key, value):
        message = pb2.Message(message=pickle.dumps((key, value)))
        resp = self.stub.PutObject(message)
        server_res = pickle.loads(resp.message)
        [res, fn_exception, fn_traceback] = server_res
        if fn_exception is not None:
            logger.error(
                f"Error putting object with key {key} on cluster: {fn_exception}."
            )
            logger.error(f"Traceback: {fn_traceback}")
            raise fn_exception
        return res

    def clear_pins(self, pins=None):
        message = pb2.Message(message=pickle.dumps(pins or []))
        self.stub.ClearPins(message)

    def run_module(self, relative_path, module_name, fn_name, fn_type, args, kwargs):
        """
        Client function to call the rpc for RunModule
        """
        # Measure the time it takes to send the message
        serialized_module = pickle.dumps(
            [relative_path, module_name, fn_name, fn_type, args, kwargs]
        )
        start = time.time()
        message = pb2.Message(message=serialized_module)
        server_res = self.stub.RunModule(message)
        end = time.time()
        logging.info(f"Time to send message: {round(end - start, 2)} seconds")
        [res, fn_exception, fn_traceback] = pickle.loads(server_res.message)
        if fn_exception is not None:
            logger.error(f"Error inside function {fn_type}: {fn_exception}.")
            logger.error(f"Traceback: {fn_traceback}")
            raise fn_exception
        return res

    def is_connected(self):
        return self._connectivity_state in [
            grpc.ChannelConnectivity.READY,
            grpc.ChannelConnectivity.IDLE,
        ]

    def shutdown(self):
        if self.channel:
            self.channel.close()



import unittest
from pathlib import Path

import runhouse as rh
import yaml

from ray import cloudpickle as pickle

S3_BUCKET = "runhouse-blob"
TEMP_LOCAL_FOLDER = Path(__file__).parents[1] / "rh-blobs"


def setup():
    from runhouse.rns.api_utils.utils import create_s3_bucket

    create_s3_bucket(S3_BUCKET)


def test_create_and_reload_local_blob_with_name():
    name = "~/my_local_blob"
    data = pickle.dumps(list(range(50)))
    my_blob = (
        rh.blob(
            data=data,
            name=name,
            system="file",
        )
        .write()
        .save()
    )

    del data
    del my_blob

    reloaded_blob = rh.Blob.from_name(name)
    reloaded_data = pickle.loads(reloaded_blob.data)
    assert reloaded_data == list(range(50))

    # Delete metadata saved locally and / or the database for the blob
    reloaded_blob.delete_configs()

    # Delete the blob
    reloaded_blob.delete_in_system()
    assert not reloaded_blob.exists_in_system()


def test_create_and_reload_local_blob_with_path():
    name = "~/my_local_blob"
    data = pickle.dumps(list(range(50)))
    my_blob = (
        rh.blob(
            data=data,
            name=name,
            path=str(TEMP_LOCAL_FOLDER / "my_blob.pickle"),
            system="file",
        )
        .write()
        .save()
    )

    del data
    del my_blob

    reloaded_blob = rh.Blob.from_name(name)
    reloaded_data = pickle.loads(reloaded_blob.data)
    assert reloaded_data == list(range(50))

    # Delete metadata saved locally and / or the database for the blob
    reloaded_blob.delete_configs()

    # Delete just the blob itself - since we define a custom path to store the blob, we want to keep the other
    # files stored in that directory
    reloaded_blob.delete_in_system()
    assert not reloaded_blob.exists_in_system()


def test_create_and_reload_anom_local_blob():
    data = pickle.dumps(list(range(50)))
    my_blob = rh.blob(
        data=data,
        system="file",
    ).write()

    reloaded_blob = rh.blob(path=my_blob.path, dryrun=True)
    reloaded_data = pickle.loads(reloaded_blob.data)
    assert reloaded_data == list(range(50))

    # Delete the blob
    reloaded_blob.delete_in_system()
    assert not reloaded_blob.exists_in_system()


def test_create_and_reload_rns_blob():
    name = "@/s3_blob"
    data = pickle.dumps(list(range(50)))
    my_blob = (
        rh.blob(
            name=name,
            data=data,
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    del data
    del my_blob

    reloaded_blob = rh.Blob.from_name(name)
    reloaded_data = pickle.loads(reloaded_blob.data)
    assert reloaded_data == list(range(50))

    # Delete metadata saved locally and / or the database for the blob and its associated folder
    reloaded_blob.delete_configs()

    # Delete the blob itself from the filesystem
    reloaded_blob.delete_in_system()
    assert not reloaded_blob.exists_in_system()


def test_create_and_reload_rns_blob_with_path():
    name = "@/s3_blob"
    data = pickle.dumps(list(range(50)))
    my_blob = (
        rh.blob(
            name=name,
            data=data,
            system="s3",
            path=f"/{S3_BUCKET}/test_blob.pickle",
            mkdir=True,
        )
        .write()
        .save()
    )

    del data
    del my_blob

    reloaded_blob = rh.Blob.from_name(name)
    reloaded_data = pickle.loads(reloaded_blob.data)
    assert reloaded_data == list(range(50))

    # Delete metadata saved locally and / or the database for the blob and its associated folder
    reloaded_blob.delete_configs()

    # Delete the blob itself from the filesystem
    reloaded_blob.delete_in_system()
    assert not reloaded_blob.exists_in_system()


def test_to_cluster_attr():
    cluster = rh.cluster(name="^rh-cpu").up_if_not()
    local_blob = rh.blob(pickle.dumps(list(range(50))), path="models/pipeline.pkl")
    cluster_blob = local_blob.to(system=cluster)
    assert isinstance(cluster_blob.system, rh.Cluster)
    assert cluster_blob._folder._fs_str == "ssh"


def test_local_to_cluster():
    name = "~/my_local_blob"
    data = pickle.dumps(list(range(50)))
    my_blob = (
        rh.blob(
            data=data,
            name=name,
            system="file",
        )
        .write()
        .save()
    )

    cluster = rh.cluster(name="^rh-cpu").up_if_not()
    my_blob = my_blob.to(system=cluster)
    blob_data = pickle.loads(my_blob.data)
    assert blob_data == list(range(50))


def test_save_blob_to_cluster():
    cluster = rh.cluster(name="^rh-cpu").up_if_not()
    # Save blob to local directory, then upload to a new "models" directory on the root path of the cluster
    rh.blob(pickle.dumps(list(range(50))), path="models/pipeline.pkl").to(
        cluster, path="models"
    )

    # Confirm the model is saved on the cluster in the `models` folder
    status_codes = cluster.run(commands=["ls models"])
    assert "pipeline.pkl" in status_codes[0][1]


def test_from_cluster():
    cluster = rh.cluster(name="^rh-cpu").up_if_not()
    config_blob = rh.blob(path="/home/ubuntu/.rh/config.yaml", system=cluster)
    config_data = yaml.safe_load(config_blob.data)
    assert len(config_data.keys()) > 4


def test_sharing_blob():
    data = pickle.dumps(list(range(50)))
    name = "shared_blob"

    my_blob = (
        rh.blob(
            data=data,
            name=name,
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    my_blob.share(
        users=["donny@run.house", "josh@run.house"],
        access_type="write",
        notify_users=False,
    )

    assert my_blob.exists_in_system()


def test_load_shared_blob():
    my_blob = rh.Blob.from_name(name="@/shared_blob")
    assert my_blob.exists_in_system()

    raw_data = my_blob.fetch()
    # NOTE: we need to do the deserialization ourselves
    assert pickle.loads(raw_data)


def test_save_anom_blob_to_s3():
    data = pickle.dumps(list(range(50)))
    my_blob = rh.blob(
        data=data,
        system="s3",
    ).write()

    assert my_blob.exists_in_system()


if __name__ == "__main__":
    setup()
    unittest.main()



import contextlib
import logging
import subprocess
from pathlib import Path
from typing import Any, Dict

import ray

import sky
import yaml
from sky.backends import backend_utils, CloudVmRayBackend

from runhouse.rh_config import configs, rns_client

from runhouse.rns.hardware.cluster import Cluster
from runhouse.rns.obj_store import _current_cluster

logger = logging.getLogger(__name__)


class OnDemandCluster(Cluster):
    RESOURCE_TYPE = "cluster"
    RECONNECT_TIMEOUT = 5

    def __init__(
        self,
        name,
        instance_type: str = None,
        num_instances: int = None,
        provider: str = None,
        dryrun=True,
        autostop_mins=None,
        use_spot=False,
        image_id=None,
        region=None,
        sky_state=None,
        **kwargs,  # We have this here to ignore extra arguments when calling from from_config
    ):
        """
        On-demand `SkyPilot <https://github.com/skypilot-org/skypilot/>`_ Cluster.

        .. note::
            To build a cluster, please use the factory method :func:`cluster`.
        """

        super().__init__(name=name, dryrun=dryrun)

        self.instance_type = instance_type
        self.num_instances = num_instances
        self.provider = provider or configs.get("default_provider")
        self.autostop_mins = (
            autostop_mins
            if autostop_mins is not None
            else configs.get("default_autostop")
        )
        self.use_spot = use_spot if use_spot is not None else configs.get("use_spot")
        self.image_id = image_id
        self.region = region

        self.address = None
        self._grpc_tunnel = None
        self.client = None
        self.sky_state = sky_state

        # Checks if state info is in local sky db, populates if so.
        status_dict = self.status(refresh=False)
        if status_dict:
            self._populate_connection_from_status_dict(status_dict)
        elif self.sky_state:
            self._save_sky_state()

        if not self.address and not dryrun:
            # Cluster status is set to INIT in the Sky DB right after starting, so we need to refresh once
            self.update_from_sky_status(dryrun=False)

    @staticmethod
    def from_config(config: dict, dryrun=False):
        return OnDemandCluster(**config, dryrun=dryrun)

    @property
    def config_for_rns(self):
        config = super().config_for_rns

        # Also store the ssh keys for the cluster in RNS
        config.update(
            {
                "instance_type": self.instance_type,
                "num_instances": self.num_instances,
                "provider": self.provider,
                "autostop_mins": self.autostop_mins,
                "use_spot": self.use_spot,
                "image_id": self.image_id,
                "region": self.region,
                "sky_state": self._get_sky_state(),
            }
        )
        return config

    def _get_sky_state(self):
        config = sky.global_user_state.get_cluster_from_name(self.name)
        if not config:
            return None
        config["status"] = config[
            "status"
        ].name  # ClusterStatus enum is not json serializable
        if config["handle"]:
            # with open(config["handle"].cluster_yaml, mode="r") as f:
            #     config["ray_config"] = yaml.safe_load(f)
            config["public_key"] = self.ssh_creds()["ssh_private_key"] + ".pub"
            config["handle"] = {
                "cluster_name": config["handle"].cluster_name,
                # This is saved as an absolute path - convert it to relative
                "cluster_yaml": self.relative_yaml_path(
                    yaml_path=config["handle"]._cluster_yaml
                ),
                "head_ip": config["handle"].head_ip,
                "launched_nodes": config["handle"].launched_nodes,
                "launched_resources": config[
                    "handle"
                ].launched_resources.to_yaml_config(),
            }
            config["handle"]["launched_resources"].pop("spot_recovery", None)

            config["ssh_creds"] = self.ssh_creds()
        return config

    def _copy_sky_yaml_from_cluster(self, abs_yaml_path: str):
        if not Path(abs_yaml_path).exists():
            Path(abs_yaml_path).parent.mkdir(parents=True, exist_ok=True)
            self.rsync("~/.sky/sky_ray.yml", abs_yaml_path, up=False)

            # Save SSH info to the ~/.ssh/config
            ray_yaml = yaml.safe_load(open(abs_yaml_path, "r"))
            backend_utils.SSHConfigHelper.add_cluster(
                self.name, [self.address], ray_yaml["auth"]
            )

    def _save_sky_state(self):
        if not self.sky_state:
            raise ValueError("No sky state to save")

        # if we're on this cluster, no need to save sky state
        current_cluster_name = _current_cluster("cluster_name")
        if self.sky_state.get("handle", {}).get("cluster_name") == current_cluster_name:
            return

        handle_info = self.sky_state.get("handle", {})

        # If we already have the cluster in local sky db,
        # we don't need to save the state, just populate the connection info from the status
        if not sky.global_user_state.get_cluster_from_name(self.name):
            # Try running a command on the cluster before saving down the state into sky db
            self.address = handle_info.get("head_ip")
            self._ssh_creds = self.sky_state["ssh_creds"]

            try:
                self.ping(timeout=self.RECONNECT_TIMEOUT)
            except TimeoutError:
                self.address = None
                self._ssh_creds = None
                print(
                    f"Timeout when trying to connect to cluster {self.name}, treating cluster as down."
                )
                return

            # TODO [JL] we need to call `ray.shutdown()` as sky does a ray init here on the cluster
            ray.shutdown()

            resources = sky.Resources.from_yaml_config(
                handle_info["launched_resources"]
            )
            # Need to convert to relative to find the yaml file in a new environment
            yaml_path = self.relative_yaml_path(handle_info.get("cluster_yaml"))
            handle = CloudVmRayBackend.ResourceHandle(
                cluster_name=self.name,
                cluster_yaml=str(Path(yaml_path).expanduser()),
                launched_nodes=handle_info["launched_nodes"],
                # head_ip=handle_info['head_ip'], # deprecated
                launched_resources=resources,
            )
            sky.global_user_state.add_or_update_cluster(
                cluster_name=self.name,
                cluster_handle=handle,
                requested_resources=[resources],
                is_launch=True,
                ready=False,
            )

        # Now try loading in the status from the sky DB
        status = self.status(refresh=False)

        abs_yaml_path = status["handle"].cluster_yaml
        try:
            if not Path(abs_yaml_path).exists():
                # This is also a good way to check if the cluster is still up
                self._copy_sky_yaml_from_cluster(abs_yaml_path)
            else:
                # We still should check if the cluster is up, since the status/yaml file could be stale
                self.ping(timeout=self.RECONNECT_TIMEOUT)
        except Exception:
            # Refresh the cluster status before saving the ssh info so SkyPilot has a chance to wipe the .ssh/config if
            # the cluster went down
            self.update_from_sky_status(dryrun=self.dryrun)

    def __getstate__(self):
        """Make sure sky_state is loaded in before pickling."""
        self.sky_state = self._get_sky_state()
        return super().__getstate__()

    @staticmethod
    def relative_yaml_path(yaml_path):
        if Path(yaml_path).is_absolute():
            yaml_path = "~/.sky/generated/" + Path(yaml_path).name
        return yaml_path

    # ----------------- Launch/Lifecycle Methods -----------------

    def is_up(self) -> bool:
        """Whether the cluster is up."""
        self.update_from_sky_status(dryrun=False)
        return self.address is not None

    def status(self, refresh: bool = True):
        """
        Get status of Sky cluster.

        Return dict looks like:

        .. code-block::

            {'name': 'sky-cpunode-donny',
             'launched_at': 1662317201,
             'handle': ResourceHandle(
                          cluster_name=sky-cpunode-donny,
                          head_ip=54.211.97.164,
                          cluster_yaml=/Users/donny/.sky/generated/sky-cpunode-donny.yml,
                          launched_resources=1x AWS(m6i.2xlarge),
                          tpu_create_script=None,
                          tpu_delete_script=None),
             'last_use': 'sky cpunode',
             'status': <ClusterStatus.UP: 'UP'>,
             'autostop': -1,
             'metadata': {}}

        .. note::
            For more information:
            https://github.com/skypilot-org/skypilot/blob/0c2b291b03abe486b521b40a3069195e56b62324/sky/backends/cloud_vm_ray_backend.py#L1457
        """  # noqa
        # return backend_utils._refresh_cluster_record(
        #     self.name, force_refresh=refresh, acquire_per_cluster_status_lock=False
        # )
        if not sky.global_user_state.get_cluster_from_name(self.name):
            return None

        state = sky.status(cluster_names=[self.name], refresh=refresh)
        # We still need to check if the cluster present in case the cluster went down and was removed from the DB
        if len(state) == 0:
            return None
        return state[0]

    def _populate_connection_from_status_dict(self, cluster_dict: Dict[str, Any]):
        if cluster_dict and cluster_dict["status"].name in ["UP", "INIT"]:
            self.address = cluster_dict["handle"].head_ip
            yaml_path = cluster_dict["handle"].cluster_yaml
            if Path(yaml_path).exists():
                self._ssh_creds = backend_utils.ssh_credential_from_yaml(yaml_path)
        else:
            self.address = None
            self._ssh_creds = None

    def update_from_sky_status(self, dryrun: bool = False):
        # Try to get the cluster status from SkyDB
        cluster_dict = self.status(refresh=not dryrun)
        self._populate_connection_from_status_dict(cluster_dict)

    def up(self):
        """Up the cluster."""
        if self.provider in ["aws", "gcp", "azure", "lambda", "cheapest"]:
            task = sky.Task(
                num_nodes=self.num_instances
                if self.instance_type and ":" not in self.instance_type
                else None,
                # docker_image=image,  # Zongheng: this is experimental, don't use it
                # envs=None,
            )
            cloud_provider = (
                sky.clouds.CLOUD_REGISTRY.from_str(self.provider)
                if self.provider != "cheapest"
                else None
            )
            task.set_resources(
                sky.Resources(
                    cloud=cloud_provider,
                    instance_type=self.instance_type
                    if self.instance_type
                    and ":" not in self.instance_type
                    and "CPU" not in self.instance_type
                    else None,
                    accelerators=self.instance_type
                    if self.instance_type
                    and ":" in self.instance_type
                    and "CPU" not in self.instance_type
                    else None,
                    cpus=self.instance_type.rsplit(":", 1)[1]
                    if self.instance_type
                    and ":" in self.instance_type
                    and "CPU" in self.instance_type
                    else None,
                    region=self.region,
                    image_id=self.image_id,
                    use_spot=self.use_spot,
                )
            )
            if Path("~/.rh").expanduser().exists():
                task.set_file_mounts(
                    {
                        "~/.rh": "~/.rh",
                    }
                )
            sky.launch(
                task,
                cluster_name=self.name,
                idle_minutes_to_autostop=self.autostop_mins,
                down=True,
            )
        elif self.provider == "k8s":
            raise NotImplementedError("Kubernetes Cluster provider not yet supported")
        else:
            raise ValueError(f"Cluster provider {self.provider} not supported.")

        self.update_from_sky_status()
        self.restart_grpc_server()

    def keep_warm(self, autostop_mins: int = -1):
        """Keep the cluster warm for given number of minutes after inactivity. If `autostop_mins` is set
        to -1, keep cluster warm indefinitely."""
        sky.autostop(self.name, autostop_mins, down=True)
        self.autostop_mins = autostop_mins

    def teardown(self):
        """Teardown cluster."""
        # Stream logs
        sky.down(self.name)
        self.address = None

    def teardown_and_delete(self):
        """Teardown cluster and delete it from configs."""
        self.teardown()
        rns_client.delete_configs()

    @contextlib.contextmanager
    def pause_autostop(self):
        sky.autostop(self.name, idle_minutes=-1)
        yield
        sky.autostop(self.name, idle_minutes=self.autostop_mins)

    # ----------------- SSH Methods ----------------- #

    @staticmethod
    def cluster_ssh_key(path_to_file):
        """Retrieve SSH key for the cluster."""
        try:
            f = open(path_to_file, "r")
            private_key = f.read()
            return private_key
        except FileNotFoundError:
            raise Exception(f"File with ssh key not found in: {path_to_file}")

    def ssh_creds(self):
        if self._ssh_creds:
            return self._ssh_creds

        if not self.status(refresh=False) and self.sky_state:
            # If this cluster was serialized and sent over the wire, it will have sky_state (we make sure of that
            # in __getstate__) but no yaml, and we need to save down the sky data to the sky db and local yaml
            self._save_sky_state()
        else:
            # To avoid calling this twice (once in save_sky_data)
            self.update_from_sky_status(dryrun=True)

        return self._ssh_creds

    def ssh(self):
        """SSH into the cluster."""
        subprocess.run(["ssh", f"{self.name}"])



import configparser
import os
from typing import Optional

from runhouse import Secrets


class AWSSecrets(Secrets):
    PROVIDER_NAME = "aws"
    CREDENTIALS_FILE = os.path.expanduser("~/.aws/credentials")

    @classmethod
    def read_secrets(cls, from_env: bool = False, file_path: Optional[str] = None):
        if from_env:
            access_key = os.getenv("AWS_ACCESS_KEY_ID")
            secret_key = os.getenv("AWS_SECRET_ACCESS_KEY")
            if not access_key or not secret_key:
                raise Exception(
                    f"AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY must be set for {cls.PROVIDER_NAME}"
                )
        else:
            if file_path:
                # Read the credentials from the given file path
                config = cls.read_config_file(file_path)
                section_name = "default"
                access_key = config[section_name]["aws_access_key_id"]
                secret_key = config[section_name]["aws_secret_access_key"]
            else:
                # TODO check if user has boto installed, if not tell them to install with runhouse[aws]
                import boto3

                session = boto3.Session()
                credentials = session.get_credentials()

                # Credentials are refreshable, so accessing your access key / secret key
                # separately can lead to a race condition.
                credentials = credentials.get_frozen_credentials()
                access_key = credentials.access_key
                secret_key = credentials.secret_key

        return {
            "access_key": access_key,
            "secret_key": secret_key,
        }

    @classmethod
    def save_secrets(cls, secrets: dict, overwrite: bool = False):
        dest_path = cls.default_credentials_path()
        cls.check_secrets_for_mismatches(
            secrets_to_save=secrets, secrets_path=dest_path, overwrite=overwrite
        )

        parser = configparser.ConfigParser()
        section_name = "default"
        parser.add_section(section_name)
        parser.set(
            section=section_name,
            option="aws_access_key_id",
            value=secrets["access_key"],
        )
        parser.set(
            section=section_name,
            option="aws_secret_access_key",
            value=secrets["secret_key"],
        )

        cls.save_to_config_file(parser, dest_path)
        cls.add_provider_to_rh_config()



import copy
import logging
import os
import shutil
import subprocess
import tempfile
from pathlib import Path
from typing import Dict, Optional, Union

import fsspec

import sshfs

import runhouse as rh
from runhouse.rh_config import rns_client
from runhouse.rns.api_utils.utils import generate_uuid
from runhouse.rns.obj_store import _current_cluster
from runhouse.rns.resource import Resource

fsspec.register_implementation("ssh", sshfs.SSHFileSystem)
# SSHFileSystem is not yet builtin.
# Line above suggested by fsspec devs: https://github.com/fsspec/filesystem_spec/issues/1071

logger = logging.getLogger(__name__)

PROVIDER_FS_LOOKUP = {
    "aws": "s3",
    "gcp": "gs",
    "azure": "abfs",
    "oracle": "ocifs",
    "databricks": "dbfs",
    "github": "github",
}


class Folder(Resource):
    RESOURCE_TYPE = "folder"
    DEFAULT_FS = "file"
    CLUSTER_FS = "ssh"
    DEFAULT_FOLDER_PATH = "/runhouse-folder"
    DEFAULT_CACHE_FOLDER = "~/.cache/runhouse"

    def __init__(
        self,
        name: Optional[str] = None,
        path: Optional[str] = None,
        system: Optional[str] = None,
        dryrun: bool = False,
        local_mount: bool = False,
        data_config: Optional[Dict] = None,
        **kwargs,  # We have this here to ignore extra arguments when calling from from_config
    ):
        """
        Runhouse Folder object.

        .. note::
            To build a folder, please use the factory method :func:`folder`.
        """
        super().__init__(name=name, dryrun=dryrun)

        self._system = None
        self._fsspec_fs = None
        self._fsspec_fs_str = None

        current_cluster_config = _current_cluster(key="config")
        if current_cluster_config and system is None:
            from runhouse.rns.hardware.cluster import Cluster

            self.system = Cluster.from_config(current_cluster_config)
        elif isinstance(system, dict):
            from runhouse.rns.hardware.cluster import Cluster

            self.system = Cluster.from_config(system)
        else:
            self.system = system or self.DEFAULT_FS

        # TODO [DG] Should we ever be allowing this to be None?
        self._path = (
            self.default_path(self.rns_address, system)
            if path is None
            else path
            if isinstance(system, Resource)
            else path
            if Path(path).expanduser().is_absolute()
            else str(Path(rns_client.locate_working_dir()) / path)
        )
        self.data_config = data_config or {}

        self.local_mount = local_mount
        self._local_mount_path = None
        if local_mount:
            self.mount(tmp=True)
        if not self.dryrun:
            self.mkdir()

    def __getstate__(self):
        """Override the pickle method to clear _fsspec_fs before pickling."""
        state = self.__dict__.copy()
        state["_fsspec_fs"] = None
        return state

    @classmethod
    def default_path(cls, rns_address, system):
        from runhouse.rns.hardware import Cluster

        if system == Folder.DEFAULT_FS or isinstance(system, Cluster):
            if rns_address:
                return str(
                    Path.cwd() / rns_client.split_rns_name_and_path(rns_address)[0]
                )  # saves to cwd / name
            return f"{Folder.DEFAULT_CACHE_FOLDER}/{generate_uuid()}"
        else:
            # If no path provided for a remote file system default to its name if provided
            if rns_address:
                name = rns_address[1:].replace("/", "_") + f".{cls.RESOURCE_TYPE}"
                return f"{Folder.DEFAULT_FOLDER_PATH}/{name}"
            return f"{Folder.DEFAULT_FOLDER_PATH}/{generate_uuid()}"

    # ----------------------------------
    @staticmethod
    def from_config(config: dict, dryrun=True):
        """Load config values into the object."""
        if config["system"] == "s3":
            from .s3_folder import S3Folder

            return S3Folder.from_config(config, dryrun=dryrun)
        elif config["system"] == "gs":
            from .gcs_folder import GCSFolder

            return GCSFolder.from_config(config, dryrun=dryrun)
        elif config["system"] == "azure":
            from .azure_folder import AzureFolder

            return AzureFolder.from_config(config, dryrun=dryrun)
        elif isinstance(config["system"], dict):
            from runhouse.rns.hardware import Cluster

            config["system"] = Cluster.from_config(config["system"], dryrun=dryrun)
        return Folder(**config, dryrun=dryrun)

    @classmethod
    def _check_for_child_configs(cls, config):
        """Overload by child resources to load any resources they hold internally."""
        system = config["system"]
        if isinstance(system, str) and rns_client.exists(system):
            # if the system is set to a cluster
            cluster_config: dict = rns_client.load_config(name=system)
            if not cluster_config:
                raise Exception(f"No cluster config saved for {system}")

            # set the cluster config as the system
            config["system"] = cluster_config
        return config

    @property
    def path(self):
        if self._path is not None:
            if self.system == Folder.DEFAULT_FS:
                return str(Path(self._path).expanduser())
            elif self._fs_str == self.CLUSTER_FS and self._path.startswith("~/"):
                # sftp takes relative paths to the home directory but doesn't understand '~'
                return str(self._path[2:])
            return str(self._path)
        else:
            return None

    @path.setter
    def path(self, path):
        self._path = path
        self._local_mount_path = None

    @property
    def system(self):
        return self._system

    @system.setter
    def system(self, data_source):
        self._system = data_source
        self._fsspec_fs = None

    # Maybe figure out how to free sshfs properly (https://github.com/ronf/asyncssh/issues/112)
    # def __del__(self):
    #     if self.local_mount:
    #         self.unmount()
    #     if self._fsspec_fs and hasattr(self._fsspec_fs, "close"):
    #         self._fsspec_fs.close()

    @property
    def data_config(self):
        if isinstance(self.system, Resource):  # if system is a cluster
            # handle case cluster is itself
            rns_address = _current_cluster("name")
            if rns_address and rns_address == self.system.rns_address:
                return self._data_config

            if not self.system.address:
                self.system.update_from_sky_status(dryrun=False)
                if not self.system.address:
                    raise ValueError(
                        "Cluster must be started before copying data from it."
                    )
            creds = self.system.ssh_creds()
            config_creds = {
                "host": self.system.address,
                "username": creds["ssh_user"],
                # 'key_filename': str(Path(creds['ssh_private_key']).expanduser())}  # For SFTP
                "client_keys": [
                    str(Path(creds["ssh_private_key"]).expanduser())
                ],  # For SSHFS
                "connect_timeout": "3s",
            }
            ret_config = self._data_config.copy()
            ret_config.update(config_creds)
            return ret_config
        return self._data_config

    @data_config.setter
    def data_config(self, data_config):
        self._data_config = data_config
        self._fsspec_fs = None

    @property
    def _fs_str(self):
        if isinstance(self.system, Resource):  # if system is a cluster
            if self.system.rns_address == _current_cluster("name"):
                return self.DEFAULT_FS
            return self.CLUSTER_FS
        else:
            return self.system

    @property
    def fsspec_fs(self):
        if self._fsspec_fs_str != self._fs_str or self._fsspec_fs is None:
            self._fsspec_fs_str = self._fs_str
            self._fsspec_fs = fsspec.filesystem(self._fsspec_fs_str, **self.data_config)
        return self._fsspec_fs

    @property
    def local_path(self):
        if self.is_local():
            return self._local_mount_path or str(Path(self.path).expanduser())
        else:
            return None

    def is_writable(self):
        # If the filesystem hasn't overridden mkdirs, it's a no-op and the filesystem is probably readonly
        # (e.g. https://filesystem-spec.readthedocs.io/en/latest/_modules/fsspec/implementations/github.html).
        # In that case, we should just create a new folder in the default
        # location and add it as a child to the parent folder.
        return self.fsspec_fs.__class__.mkdirs == fsspec.AbstractFileSystem.mkdirs

    def mv(
        self, system, path: Optional[str] = None, data_config: Optional[dict] = None
    ) -> None:
        """Move the folder to a new filesystem."""
        # TODO [DG] create get_default_path for system method to be shared
        if path is None:
            path = "rh/" + self.rns_address
        data_config = data_config or {}
        with fsspec.open(self.fsspec_url, **self.data_config) as src:
            with fsspec.open(f"{system}://{path}", **data_config) as dest:
                # NOTE For packages, maybe use the `ignore` param here to only copy python files.
                new_path = shutil.move(src, dest)
        self.path = new_path
        self.system = system
        self.data_config = data_config or {}

    def to(
        self,
        system: Union[str, "Cluster"],
        path: Optional[str] = None,
        data_config: Optional[dict] = None,
    ):
        """Copy the folder to a new filesystem, and return a new Folder object pointing to the new location."""
        if system == "here":
            current_cluster_config = _current_cluster(key="config")
            if current_cluster_config:
                from runhouse.rns.hardware.cluster import Cluster

                system = Cluster.from_config(current_cluster_config)
            else:
                system = "file"
            path = str(Path.cwd() / self.path.split("/")[-1]) if path is None else path

        path = str(
            path or self.default_path(self.rns_address, system)
        )  # Make sure it's a string and not a Path

        system_str = getattr(
            system, "name", system
        )  # Use system.name if available, i.e. system is a cluster
        logging.info(
            f"Copying folder from {self.fsspec_url} to: {system_str}, with path: {path}"
        )

        # to_local, to_cluster and to_data_store are also overridden by subclasses to dispatch
        # to more performant cloud-specific APIs
        from runhouse.rns.hardware import Cluster

        if isinstance(system, str) and rns_client.exists(
            system, resource_type="cluster"
        ):
            system = Cluster.from_name(system, dryrun=self.dryrun)

        if system == "file":
            return self.to_local(dest_path=path, data_config=data_config)
        elif isinstance(system, Cluster):  # If system is a cluster
            return self.to_cluster(dest_cluster=system, path=path)
        elif system in ["s3", "gs", "azure"]:
            return self.to_data_store(
                system=system, data_store_path=path, data_config=data_config
            )
        else:
            self.fsspec_copy(system, path, data_config)
            new_folder = copy.deepcopy(self)
            new_folder.path = path
            new_folder.system = system
            new_folder.data_config = data_config or {}
            return new_folder

    def fsspec_copy(self, system: str, path: str, data_config: dict):
        """Copy the fsspec filesystem to the given new filesystem and path."""
        # Fallback for other fsspec filesystems, but very slow:
        if self.is_local():
            self.fsspec_fs.put(self.path, f"{system}://{path}", recursive=True)
        else:
            # This is really really slow, maybe use skyplane, as follows:
            # src_url = f'local://{self.path}' if self.is_local() else self.fsspec_url
            # subprocess.run(['skyplane', 'sync', src_url, f'{system}://{path}'])

            # FYI: from https://github.com/fsspec/filesystem_spec/issues/909
            # Maybe copy chunks https://github.com/fsspec/filesystem_spec/issues/909#issuecomment-1204212507
            src = fsspec.get_mapper(self.fsspec_url, create=False, **self.data_config)
            dest = fsspec.get_mapper(f"{system}://{path}", create=True, **data_config)
            # dest.system.mkdir(dest.root, create_parents=True)
            import tqdm

            for k in tqdm.tqdm(src):
                # NOTE For packages, maybe use the `ignore` param here to only copy python files.
                dest[k] = src[k]
                # dst.write(src.read())

    def destination_folder(
        self,
        dest_path: str,
        dest_system: Optional[str] = "file",
        data_config: Optional[dict] = None,
    ):
        """Returns a new Folder object pointing to the destination folder."""
        new_folder = copy.deepcopy(self)
        new_folder.path = dest_path
        new_folder.system = dest_system
        new_folder.data_config = data_config or {}
        return new_folder

    def to_local(self, dest_path: str, data_config: dict):
        """Copies folder to local."""
        from runhouse.rns.hardware import Cluster

        if (
            self._fs_str == "file"
        ):  # Also covers the case where we're on the cluster at system
            # Simply move the files within local system
            shutil.copytree(src=self.path, dst=dest_path)
        elif isinstance(self.system, Cluster):
            return self._cluster_to_local(cluster=self.system, dest_path=dest_path)
        else:
            self.fsspec_copy("file", dest_path, data_config)

        return self.destination_folder(
            dest_path=dest_path, dest_system="file", data_config=data_config
        )

    def to_data_store(
        self,
        system: str,
        data_store_path: Optional[str] = None,
        data_config: Optional[dict] = None,
    ):
        """Local or cluster to blob storage."""
        from runhouse.rns.hardware import Cluster

        local_folder_path = self.path

        folder_config = self.config_for_rns
        folder_config["system"] = system
        folder_config["path"] = data_store_path
        folder_config["data_config"] = data_config
        new_folder = Folder.from_config(folder_config)

        if (
            self._fs_str == "file"
        ):  # Also covers the case where we're on the cluster at system
            new_folder.upload(src=local_folder_path)
        elif isinstance(self.system, Cluster):
            self.system.run(
                [new_folder.upload_command(src=local_folder_path, dest=new_folder.path)]
            )
        else:
            self.fsspec_copy("file", data_store_path, data_config)

        return new_folder

    @staticmethod
    def rsync(local, remote, data_config, up=True):
        """Rsync local folder to remote."""
        dest_str = f'{data_config["username"]}@{data_config["host"]}:{remote}'
        src_str = local
        if not up:
            src_str, dest_str = dest_str, src_str
        subprocess.check_call(
            f"rsync {src_str} {dest_str} "
            f'--password_file {data_config["key_filename"]}'
        )

    def mkdir(self):
        """Create the folder in specified file system if it doesn't already exist."""
        folder_path = self.path
        if Path(os.path.basename(folder_path)).suffix != "":
            folder_path = str(Path(folder_path).parent)

        logging.info(
            f"Creating new {self._fs_str} folder if it does not already exist in path: {folder_path}"
        )
        self.fsspec_fs.mkdirs(folder_path, exist_ok=True)

    def mount(self, path: Optional[str] = None, tmp: bool = False) -> str:
        """Mount the folder locally."""
        # TODO check that fusepy and FUSE are installed
        if tmp:
            self._local_mount_path = tempfile.mkdtemp()
        else:
            self._local_mount_path = path
        remote_fs = self.fsspec_fs
        fsspec.fuse.run(
            fs=remote_fs, path=self.path, mount_point=self._local_mount_path
        )
        return self._local_mount_path

    def to_cluster(self, dest_cluster, path=None, mount=False):
        """Copy the folder from a file or cluster source onto a destination cluster."""
        if not dest_cluster.address:
            raise ValueError("Cluster must be started before copying data to it.")

        # Create tmp_mount if needed
        if not self.is_local() and mount:
            self.mount(tmp=True)

        dest_path = path or f"~/{Path(self.path).name}"

        # Need to add slash for rsync to copy the contents of the folder
        dest_folder = copy.deepcopy(self)
        dest_folder.path = dest_path
        dest_folder.system = dest_cluster

        if self._fs_str == "file":  # Includes case where we're on the cluster
            dest_cluster.rsync(source=self.path, dest=dest_path, up=True, contents=True)

        elif isinstance(self.system, Resource):
            src_path = self.path

            cluster_creds = self.system.ssh_creds()
            creds_file = cluster_creds["ssh_private_key"]

            dest_cluster.run([f"mkdir -p {dest_path}"])
            command = (
                f"rsync -Pavz --filter='dir-merge,- .gitignore' -e \"ssh -i '{creds_file}' "
                f"-o StrictHostKeyChecking=no -o IdentitiesOnly=yes -o ExitOnForwardFailure=yes "
                f"-o ServerAliveInterval=5 -o ServerAliveCountMax=3 -o ConnectTimeout=30s -o ForwardAgent=yes "
                f'-o ControlMaster=auto -o ControlPersist=300s" {src_path}/ {dest_cluster.address}:{dest_path}'
            )
            status_codes = self.system.run([command])
            if status_codes[0][0] != 0:
                raise Exception(
                    f"Error syncing folder to destination cluster ({dest_cluster.name}). "
                    f"Make sure the source cluster ({self.system.name}) has the necessary provider keys "
                    f"loaded in path: {creds_file}. "
                    f"For example: `rh.Secrets.to({self.system.name}, providers=['aws'])`"
                )
        else:
            raise TypeError(
                f"`to_cluster` not supported for filesystem type {type(self.system)}"
            )

        return dest_folder

    def _cluster_to_local(self, cluster, dest_path):
        """Create a local folder with dest_path from the cluster.

        This function rsyncs down the data and return a folder with system=='file'.
        """
        if not cluster.address:
            raise ValueError("Cluster must be started before copying data from it.")
        Path(dest_path).expanduser().mkdir(parents=True, exist_ok=True)
        cluster.rsync(
            source=self.path,
            dest=str(Path(dest_path).expanduser()),
            up=False,
            contents=True,
        )
        new_folder = copy.deepcopy(self)
        new_folder.path = dest_path
        new_folder.system = "file"
        # Don't need to do anything with _data_config because cluster creds are injected virtually through the
        # data_config property
        return new_folder

    def is_local(self):
        """Whether the folder is on the local filesystem."""
        return (
            self._fs_str == "file"
            and self.path is not None
            and Path(self.path).exists()
        ) or self._local_mount_path

    def upload(self, src: str, region: Optional[str] = None):
        """Upload a folder to a remote bucket."""
        raise NotImplementedError

    def upload_command(self, src: str, dest: str):
        """CLI command for uploading folder to remote bucket. Needed when uploading a folder from a cluster."""
        raise NotImplementedError

    def run_upload_cli_cmd(self, sync_dir_command: str, access_denied_message: str):
        """Uploads a folder to a remote bucket.
        Based on the CLI command skypilot uses to upload the folder"""
        from sky.data.data_utils import run_upload_cli

        run_upload_cli(
            command=sync_dir_command,
            access_denied_message=access_denied_message,
            bucket_name=self.bucket_name_from_path(self.path),
        )

    def download(self, dest):
        raise NotImplementedError

    def download_command(self, src, dest):
        """CLI command for downloading folder from remote bucket. Needed when downloading a folder to a cluster."""
        raise NotImplementedError

    @property
    def config_for_rns(self):
        config = super().config_for_rns
        config_attrs = ["local_mount", "data_config"]
        self.save_attrs_to_config(config, config_attrs)

        if self.system == Folder.DEFAULT_FS:
            # If folder is local check whether path is relative, and if so take it relative to the working director
            # rather than to the home directory. If absolute, it's left alone.
            config["path"] = (
                self._path_relative_to_rh_workdir(self.path) if self.path else None
            )
        else:
            # if not a local filesystem save path as is (e.g. bucket/path)
            config["path"] = self.path

        if isinstance(self.system, Resource):  # If system is a cluster
            config["system"] = self._resource_string_for_subconfig(self.system)
        else:
            config["system"] = self.system

        return config

    def _save_sub_resources(self):
        if isinstance(self.system, Resource):
            self.system.save()

    @staticmethod
    def _path_relative_to_rh_workdir(path):
        rh_workdir = Path(rns_client.locate_working_dir())
        try:
            return str(Path(path).relative_to(rh_workdir))
        except ValueError:
            return path

    @property
    def fsspec_url(self):
        """Generate the FSSpec URL using the file system and path of the folder"""
        if self.path.startswith("/") and self._fs_str not in [
            rns_client.DEFAULT_FS,
            self.CLUSTER_FS,
        ]:
            return f"{self._fs_str}:/{self.path}"
        else:
            # For local, ssh / sftp filesystems we need both slashes
            # e.g.: 'ssh:///home/ubuntu/.cache/runhouse/tables/dede71ef83ce45ffa8cb27d746f97ee8'
            return f"{self._fs_str}://{self.path}"

    def ls(self, full_paths: bool = True):
        """List the contents of the folder"""
        paths = self.fsspec_fs.ls(path=self.path) if self.path else []
        if full_paths:
            return paths
        else:
            return [Path(path).name for path in paths]

    def resources(self, full_paths: bool = False, resource_type: str = None):
        """List the resources in the *RNS* folder."""
        # TODO allow '*' wildcard for listing all resources (and maybe other wildcard things)
        try:
            resources = [
                path for path in self.ls() if (Path(path) / "config.json").exists()
            ]
        except FileNotFoundError:
            return []

        if full_paths:
            return [self.rns_address + "/" + Path(path).stem for path in resources]
        else:
            return [Path(path).stem for path in resources]

    @property
    def rns_address(self):
        """Traverse up the filesystem until reaching one of the directories in rns_base_folders,
        then compute the relative path to that.
        """
        # TODO Maybe later, account for folders along the path with a different RNS name.

        if self.name is None:  # Anonymous folders have no rns address
            return None

        # Only should be necessary when a new base folder is being added (therefore isn't in rns_base_folders yet)
        if self._rns_folder:
            return str(Path(self._rns_folder) / self.name)

        if self.path in rns_client.rns_base_folders.values():
            if self._rns_folder:
                return self._rns_folder + "/" + self.name
            else:
                return rns_client.default_folder + "/" + self.name

        segment = Path(self.path)
        while (
            str(segment) not in rns_client.rns_base_folders.values()
            and not segment == Path.home()
            and not segment == segment.parent
        ):
            segment = segment.parent

        if (
            segment == Path.home() or segment == segment.parent
        ):  # TODO throw an error instead?
            return rns_client.default_folder + "/" + self.name
        else:
            base_folder = Folder(path=str(segment), dryrun=True)
            base_folder_path = base_folder.rns_address
            relative_path = str(Path(self.path).relative_to(base_folder.path))
            return base_folder_path + "/" + relative_path

    def contains(self, name_or_path) -> bool:
        """Whether path of a Folder exists locally."""
        path, system = self.locate(name_or_path)
        return path is not None

    def locate(self, name_or_path) -> (str, str):
        """Locate the local path of a Folder given an rns path."""
        # Note: Keep in mind we're using both _rns_ path and physical path logic below. Be careful!

        # If the path is already given relative to the current folder:
        if (Path(self.path) / name_or_path).exists():
            return str(Path(self.path) / name_or_path), self.system

        # If name or path uses ~/ or ./, need to resolve with folder path
        abs_path = rns_client.resolve_rns_path(name_or_path)
        rns_path = self.rns_address

        # If this folder is anonymous, it has no rns contents
        if rns_path is None:
            return None, None

        if abs_path == rns_path:
            return self.path, self.system
        try:
            child_path = Path(self.path) / Path(abs_path).relative_to(rns_path)
            if child_path.exists():
                return str(child_path), self.system
        except ValueError:
            pass

        # Last resort, recursively search inside sub-folders and children.

        segments = abs_path.lstrip("/").split("/")
        if len(segments) == 1:
            return (
                None,
                None,
            )  # If only a single element, would have been found in ls above.

        # Look for lowest folder in the path that exists in filesystem, and recurse from that folder
        greatest_common_folder = Path(self.path)
        i = 0
        for i, seg in enumerate(segments):
            if not (greatest_common_folder / seg).exists():
                break
            greatest_common_folder = greatest_common_folder / seg
        if not str(greatest_common_folder) == self.path:
            return Folder(
                path=str(greatest_common_folder), system=self.system, dryrun=True
            ).locate("/".join(segments[i + 1 :]))

        return None, None

    def open(self, name, mode="rb", encoding=None):
        """Returns an fsspec file, which must be used as a content manager to be opened.

        Example:
            >>> with my_folder.open('obj_name') as my_file:
            >>>        pickle.load(my_file)
        """
        return self.fsspec_fs.open(self.path + "/" + name, mode=mode, encoding=encoding)

    def get(self, name, mode="rb", encoding=None):
        """Returns the contents of a file as a string or bytes."""
        with self.open(name, mode=mode, encoding=encoding) as f:
            return f.read()

    # TODO [DG] fix this to follow the correct convention above
    def get_all(self):
        # TODO add docs for this
        # TODO we're not closing these, do we need to extract file-like objects so we can close them?
        return fsspec.open_files(self.fsspec_url, mode="rb", **self.data_config)

    def exists_in_system(self):
        """Whether the folder exists in the filesystem."""
        return self.fsspec_fs.exists(
            self.fsspec_url
        ) or rh.rns.top_level_rns_fns.exists(self.path)

    def delete_in_system(self):
        """Delete from file system."""
        try:
            self.fsspec_fs.rmdir(self.path)
        except FileNotFoundError:
            pass

    def rm(self, name, recursive: bool = True):
        """Remove a resource from the folder."""
        try:
            self.fsspec_fs.rm(self.fsspec_url + "/" + name, recursive=recursive)
        except FileNotFoundError:
            pass

    def put(self, contents, overwrite=False):
        """Put given contents in folder. Contents must be one of the following:

        - Dict with keys being the file names and values being the file-like objects to write

        - Resource

        - List of Resources.
        """
        # TODO create the bucket if it doesn't already exist
        # Handle lists of resources just for convenience
        if isinstance(contents, list):
            for resource in contents:
                self.put(resource, overwrite=overwrite)
            return

        if isinstance(contents, Folder):
            if not self.is_writable():
                raise RuntimeError(
                    f"Cannot put files into non-writable folder {self.name or self.path}"
                )
            if contents.path is None:  # Should only be the case when Folder is created
                contents.path = self.path + "/" + contents.name
                contents.system = self.system
                # The parent can be anonymous, e.g. the 'rh' folder.
                # TODO not sure if this should be allowed - if parent folder has no rns address, why would child
                # just be put into the default rns folder?
                # TODO If the base is named later, figure out what to do with the contents (rename, resave, etc.).
                if self.rns_address is None:
                    contents.rns_path = rns_client.default_folder + "/" + contents.name
                    rns_client.rns_base_folders.update(
                        {contents.rns_address: contents.path}
                    )
                # We don't need to call .save here to write down because it will be called at the end of the
                # folder or resource constructor
            else:
                if contents.name is None:  # Anonymous resource
                    i = 1
                    new_name = contents.RESOURCE_TYPE + str(i)
                    # Resolve naming conflicts if necessary
                    while rns_client.exists(self.path + "/" + new_name):
                        i += 1
                        new_name = contents.RESOURCE_TYPE + str(i)
                else:
                    new_name = contents.name

                # NOTE For intercloud transfer, we should use Skyplane
                with fsspec.open(
                    self.fsspec_url + "/" + new_name, **self.data_config
                ) as dest:
                    with fsspec.open(
                        contents.fsspec_url, **contents.data_config
                    ) as src:
                        # NOTE For packages, maybe use the `ignore` param here to only copy python files.
                        shutil.move(src, dest)
            return

        if not isinstance(contents, dict):
            raise TypeError(
                "`files` argument to `.put` must be Resource, list of Resources, or dict mapping "
                "filenames to file-like-objects"
            )

        if overwrite is False:
            folder_contents = self.resources()
            intersection = set(folder_contents).intersection(set(contents.keys()))
            if intersection != set():
                raise FileExistsError(
                    f"File(s) {intersection} already exist(s) at path"
                    f"{self.path}, cannot save them without overwriting."
                )
        # TODO figure out default behavior for not overwriting but still saving
        # if not overwrite:
        #     time = datetime.today().strftime('%Y-%m-%d_%H:%M:%S')
        #     self.data_url = self.data_url + time or time
        filenames = list(contents)
        fss_files = fsspec.open_files(
            self.fsspec_url + "/*",
            mode="wb",
            **self.data_config,
            num=len(contents),
            name_function=filenames.__getitem__,
        )
        for (fss_file, raw_file) in zip(fss_files, contents.values()):
            with fss_file as f:
                f.write(raw_file)

    @staticmethod
    def bucket_name_from_path(path: str) -> str:
        """Extract the bucket name from a path (e.g. '/my-bucket/my-folder/my-file.txt' -> 'my-bucket')"""
        return Path(path).parts[1]


def folder(
    name: Optional[str] = None,
    path: Optional[Union[str, Path]] = None,
    system: Optional[str] = None,
    dryrun: bool = False,
    local_mount: bool = False,
    data_config: Optional[Dict] = None,
    load: bool = True,
) -> Folder:
    """Creates a Runhouse folder object, which can be used to interact with the folder at the given path.

    Args:
        name (Optional[str]): Name to give the folder, to be re-used later on.
        path (Optional[str or Path]): Path (or path) that the folder is located at.
        system (Optional[str]): File system. Currently this must be one of:
            [``file``, ``github``, ``sftp``, ``ssh``,``s3``, ``gs``, ``azure``].
            We are working to add additional file system support.
        dryrun (bool): Whether to create the Folder if it doesn't exist, or load a Folder object as a dryrun.
            (Default: ``False``)
        local_mount (bool): Whether or not to mount the folder locally. (Default: ``False``)
        data_config (Optional[Dict]): The data config to pass to the underlying fsspec handler.
        load (bool): Whether to load an existing config for the Folder. (Default: ``True``)

    Returns:
        Folder: The resulting folder.

    Example:
        >>> rh.folder(name='training_imgs', path='remote_directory/images', system='s3')
    """
    # TODO [DG] Include loud warning that relative paths are relative to the git root / working directory!

    config = rns_client.load_config(name) if load else {}
    config["name"] = name or config.get("rns_address", None) or config.get("name")
    config["path"] = path or config.get("path")
    config["local_mount"] = local_mount or config.get("local_mount")
    config["data_config"] = data_config or config.get("data_config")

    file_system = system or config.get("system") or Folder.DEFAULT_FS
    config["system"] = file_system
    if isinstance(file_system, str):
        if file_system in ["file", "github", "sftp", "ssh"]:
            new_folder = Folder.from_config(config, dryrun=dryrun)
        elif file_system == "s3":
            from .s3_folder import S3Folder

            new_folder = S3Folder.from_config(config, dryrun=dryrun)
        elif file_system == "gs":
            from .gcs_folder import GCSFolder

            new_folder = GCSFolder.from_config(config, dryrun=dryrun)
        elif file_system == "azure":
            from .azure_folder import AzureFolder

            new_folder = AzureFolder.from_config(config, dryrun=dryrun)
        elif file_system in fsspec.available_protocols():
            logger.warning(
                f"fsspec file system {file_system} not officially supported. Use at your own risk."
            )
            new_folder = Folder.from_config(config, dryrun=dryrun)
        elif rns_client.exists(file_system, resource_type="cluster"):
            config["system"] = rns_client.load_config(file_system)
        else:
            raise ValueError(
                f"File system {file_system} not found. Have you installed the "
                f"necessary packages for this fsspec protocol? (e.g. s3fs for s3). If the file system "
                f"is a cluster (ex: /my-user/rh-cpu), make sure the cluster config has been saved."
            )

    # If cluster is passed as the system.
    if isinstance(config["system"], dict) or isinstance(
        config["system"], Resource
    ):  # if system is a cluster
        new_folder = Folder.from_config(config, dryrun=dryrun)

    return new_folder
